# -*- coding: utf-8 -*-
# pyknic/lib/bellboy/app.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing
from abc import ABCMeta

import aiohttp
import pydantic

from pyknic.lib.bellboy.models import SecretBackendType
from pyknic.lib.bellboy.secret_backend import SecretBackend, KeyringSecretBackendImplementation, SecretTokenModel
from pyknic.lib.bellboy.secret_backend import SharedMemorySecretBackend, SecretBackendImplementationProto
from pyknic.lib.fastapi.lobby import LobbyCommandHandler
from pyknic.lib.fastapi.lobby_fingerprint import LobbyFingerprint
from pyknic.lib.fastapi.headers import FastAPIHeaders
from pyknic.lib.fastapi.models.lobby import LobbyCommandRequest, LobbyFingerprintModel, LobbyCommandResult
from pyknic.lib.registry import APIRegistry, register_api

__default_bellboy_commands_registry__ = APIRegistry()  # default registry for all commands


def register_bellboy_command(
    registry: APIRegistry | None = None,
) -> typing.Callable[..., typing.Any]:
    """This decorator help to register commands with the given registry."""

    if registry is None:
        registry = __default_bellboy_commands_registry__

    return register_api(
        registry=registry,
        api_id=lambda x: x.command_name(),
        callable_api_id=True
    )


class BellboyCLIError(Exception):
    """Indicates errors in Bellboy CLI."""
    pass


class BellBoyCommandHandler(LobbyCommandHandler, metaclass=ABCMeta):
    """ This is an extended version of the :class:`.LobbyCommandHandler`
    """

    @classmethod
    def secret_backend(cls, secret_backend_type: SecretBackendType) -> SecretBackend:
        """ Return a backend implementation by a specified type

        :param secret_backend_type: type of secret backend
        """
        if secret_backend_type == SecretBackendType.keyring:
            secret_backend: SecretBackendImplementationProto = KeyringSecretBackendImplementation()
        elif secret_backend_type == SecretBackendType.shm:
            secret_backend = SharedMemorySecretBackend()
        else:
            raise BellboyCLIError(f'Unknown backend spotted -- {secret_backend_type}')

        return SecretBackend(backend_implementation=secret_backend)

    @classmethod
    def auth_data(cls, secret_backend_type: SecretBackendType, lobby_url: str) -> SecretTokenModel:
        """ Return an auth info for the specified URL from the specified secret backend

        :param secret_backend_type: backend that holds secret
        :param lobby_url: URL which secret should be retrieved
        """
        secret_backend = cls.secret_backend(secret_backend_type)
        all_secrets = secret_backend.get_secrets()

        if lobby_url not in all_secrets.secrets:
            raise BellboyCLIError('Login at first')

        return all_secrets.secrets[lobby_url]


class LobbyClient:
    """This class wraps API calls routine.
    """

    def __init__(self, url: str, fingerprint: typing.Union[LobbyFingerprint | LobbyFingerprintModel], token: str):
        """Create a new client.

        :param url: URL to connect to.
        :param fingerprint: allowed server's fingerprint to check a response
        :param token: token to authenticate with.
        """
        is_lobby_fingerprint = isinstance(fingerprint, LobbyFingerprint)

        self.__url = url
        self.__fingerprint = fingerprint if is_lobby_fingerprint else LobbyFingerprint.from_model(fingerprint)
        self.__token = token

    async def secure_request(
        self,
        fingerprint: LobbyFingerprint,
        session: aiohttp.ClientSession,
        method_name: str,
        path: str | None = None,
        data: str | bytes | None = None
    ) -> bytes:
        """Make a secure request and return bytes that this request returns.

        :param fingerprint: allowed server's fingerprint to check a response
        :param session: HTTP-client to use
        :param method_name: HTTP-method to use (like 'get' or 'post')
        :param path: URL path
        :param data: Data to send with request
        """
        auth_headers = {"Authorization": f"Bearer {self.__token}"}

        session_method = getattr(session, method_name)
        async with session_method(f'{self.__url}{path if path else ""}', headers=auth_headers, data=data) as response:
            if response.status != 200:
                raise BellboyCLIError(f'API request failed with status code {response.status}')

            binary_body = await response.content.read()
            signature = fingerprint.sign(binary_body, encode_base64=True).decode('ascii')

            if response.headers[FastAPIHeaders.fingerprint.value] != signature:
                raise BellboyCLIError(
                    'Response signature mismatch! Consider to restart session and validate connectivity'
                )
            return binary_body  # type: ignore[no-any-return]

    async def command_request(
        self, command: LobbyCommandRequest, session: typing.Optional[aiohttp.ClientSession] = None
    ) -> LobbyCommandResult:
        """Send command to a server.
        :param command: Command to send to a server
        :param session: HTTP-client to use (a new session will be made if this parameter is omitted)
        """

        async def cmd_request(s: aiohttp.ClientSession) -> LobbyCommandResult:
            request_json = command.model_dump_json()
            result = await self.secure_request(
                self.__fingerprint, s, 'post', data=request_json  # type: ignore[arg-type]
            )
            return pydantic.TypeAdapter(LobbyCommandResult).validate_json(result.decode())

        if session:
            return await cmd_request(session)

        async with aiohttp.ClientSession() as new_session:
            return await cmd_request(new_session)

    @staticmethod
    async def fingerprint(url: str, session: typing.Optional[aiohttp.ClientSession] = None) -> LobbyFingerprint:
        """Return server's fingerprint.

        :param url: URL to connect to.
        :param session: HTTP-client to use
        """
        # TODO: persist it from a server side

        async def fp_by_session(s: aiohttp.ClientSession) -> LobbyFingerprint:
            async with s.get(f'{url}/fingerprint') as response:

                if response.status != 200:
                    raise BellboyCLIError('Unable to fetch fingerprint')

                fingerprint_model = LobbyFingerprintModel.model_validate(await response.json())
                return LobbyFingerprint.deserialize(fingerprint_model.fingerprint.encode('ascii'))

        if session is not None:
            return await fp_by_session(session)

        async with aiohttp.ClientSession() as new_session:
            return await fp_by_session(new_session)



# -*- coding: utf-8 -*-

import typing
import pytest

from pyknic.lib.signals.proto import Signal, SignalSourceProto, UnknownSignalException
from pyknic.lib.signals.source import SignalSourceMeta, SignalSource

from pyknic.lib.x_mansion import CapabilitiesAndSignals


class TestSignalSourceMeta:

    def test(self) -> None:

        class A(metaclass=SignalSourceMeta):
            pass

        assert(A.__pyknic_signals__ == set())  # type: ignore[attr-defined] # metaclass and mypy issues

        class B(metaclass=SignalSourceMeta):
            signal1 = Signal()
            signal2 = Signal()

        assert(B.__pyknic_signals__ == {B.signal1, B.signal2})  # type: ignore[attr-defined] # metaclass and mypy issues
        assert(B.signal1.__pyknic_signal_name__ == 'B.signal1')
        assert(B.signal2.__pyknic_signal_name__ == 'B.signal2')

        class C(B):
            signal3 = Signal()

        assert(C.signal1.__pyknic_signal_name__ == 'B.signal1')
        assert(C.signal2.__pyknic_signal_name__ == 'B.signal2')

        with pytest.raises(TypeError):
            class D(B):
                signal1 = Signal()  # signals may not be overridden


class TestSignalSource:

    @pytest.mark.parametrize(
        "test_cls", [
            SignalSource,
            CapabilitiesAndSignals,
        ]
    )
    def test(self, test_cls: typing.Type[SignalSource]) -> None:
        source = test_cls()
        assert(isinstance(source, SignalSource) is True)
        assert(isinstance(source, SignalSourceProto) is True)

        pytest.raises(UnknownSignalException, source.emit, Signal())
        pytest.raises(UnknownSignalException, source.callback, Signal(), lambda: None)
        pytest.raises(UnknownSignalException, source.remove_callback, Signal(), lambda: None)

    @pytest.mark.parametrize(
        "test_cls", [
            SignalSource,
            CapabilitiesAndSignals,
        ]
    )
    def test_emit(self, test_cls: typing.Type[SignalSource]) -> None:

        class Source(test_cls):  # type: ignore[valid-type, misc]  # mypy issues will be fixed in future releases
            signal1 = Signal()
            signal2 = Signal(int)

        results = []

        def callback(source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
            nonlocal results  # noqa: F824
            results.append((source, signal, value))

        s = Source()
        s.callback(Source.signal1, callback)
        assert(results == [])
        s.emit(Source.signal1)
        s.emit(Source.signal2, 1)
        assert(results == [(s, Source.signal1, None)])

        pytest.raises(TypeError, s.emit, Source.signal2, 'foo')

        results.clear()
        s.callback(Source.signal2, callback)
        s.emit(Source.signal1)
        s.emit(Source.signal2, 1)
        assert(results == [(s, Source.signal1, None), (s, Source.signal2, 1)])

        results.clear()
        s.remove_callback(Source.signal1, callback)
        s.emit(Source.signal1)
        s.emit(Source.signal2, 1)
        assert(results == [(s, Source.signal2, 1)])

    @pytest.mark.parametrize(
        "test_cls", [
            SignalSource,
            CapabilitiesAndSignals,
        ]
    )
    def test_bounded_callbacks(self, test_cls: typing.Type[SignalSource]) -> None:

        class Source(test_cls):  # type: ignore[valid-type, misc]  # mypy issues will be fixed in future releases
            signal1 = Signal()

        class A:
            def callback(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                pass

            @classmethod
            def cls_callback(cls, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                pass

            def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                pass

        s = Source()
        with pytest.raises(ValueError):
            s.callback(Source.signal1, A().callback)

        s.callback(Source.signal1, A.cls_callback)  # classmethods are ok
        s.callback(Source.signal1, A())  # callable objects are ok too



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/record.py
#
# Copyright (C) 2017-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.tasks.proto import ScheduleRecordProto, TaskProto, ScheduledTaskPostponePolicy, ScheduleSourceProto


class ScheduleRecord(ScheduleRecordProto):
    """ The :class:`.ScheduleRecordProto` implementation. Implementation is pretty straightforward
    """

    def __init__(
        self,
        task: TaskProto,
        source: ScheduleSourceProto,
        group_id: typing.Optional[str] = None,
        ttl: typing.Optional[typing.Union[int, float]] = None,
        simultaneous_runs: typing.Optional[int] = None,
        postpone_policy: typing.Optional[ScheduledTaskPostponePolicy] = None
    ):
        """ Create a new record

        :param task: a task to execute
        :param group_id: record's group id. Please read :meth:`.ScheduleRecordProto.group_id`
        :param ttl: record's ttl
        :param simultaneous_runs: how many records with the same group_id should be running at the moment
        :param postpone_policy: what should be done if this record will be postponed
        """
        self.__task = task
        self.__source = source
        self.__group_id = group_id
        self.__ttl = ttl
        self.__simultaneous_runs = simultaneous_runs if simultaneous_runs is not None else 0
        self.__postpone_policy = postpone_policy if postpone_policy else ScheduledTaskPostponePolicy.wait

    def task(self) -> TaskProto:
        """ The :meth:`.ScheduleRecordProto.task` method implementation
        """
        return self.__task

    def source(self) -> 'ScheduleSourceProto':
        """ The :meth:`.ScheduleRecordProto.source` method implementation
        """
        return self.__source

    def group_id(self) -> typing.Optional[str]:
        """ The :meth:`.ScheduleRecordProto.group_id` method implementation
        """
        return self.__group_id

    def ttl(self) -> typing.Union[int, float, None]:
        """ The :meth:`.ScheduleRecordProto.ttl` method implementation
        """
        return self.__ttl

    def simultaneous_runs(self) -> int:
        """ The :meth:`.ScheduleRecordProto.simultaneous_policy` method implementation
        """
        return self.__simultaneous_runs

    def postpone_policy(self) -> ScheduledTaskPostponePolicy:
        """ The :meth:`.ScheduleRecordProto.postpone_policy` method implementation
        """
        return self.__postpone_policy



# -*- coding: utf-8 -*-
# pyknic/lib/io/clients/sftp.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import typing

import paramiko

from pyknic.lib.registry import register_api
from pyknic.lib.uri import URI, URIQuery
from pyknic.lib.io.clients.virtual_dir import VirtualDirectoryClient, path_to_str
from pyknic.lib.io.clients.collection import __default_io_clients_registry__
from pyknic.lib.verify import verify_value


class _SFTPSyncImplementation:
    """This is a blocking (synchronous) SFTP client implementation.
    """

    @verify_value(uri=lambda x: x.hostname is not None)
    def __init__(self, uri: URI, vd_client: VirtualDirectoryClient):
        """Create a new SFTP client.

        :param uri: URI to connect with
        :param vd_client: original async client
        """

        self.__uri = uri
        self.__vd_client = vd_client
        self.__ssh_client: typing.Optional[paramiko.SSHClient] = None
        self.__sftp_client: typing.Optional[paramiko.SFTPClient] = None

        self.__host_key_auto_add = False
        self.__ssh_agent = False
        self.__private_key_file = None

        if self.__uri.query is not None:
            uri_query = URIQuery.parse(self.__uri.query)

            if 'host_key_auto_add' in uri_query:
                self.__host_key_auto_add = uri_query.single_parameter('host_key_auto_add', bool)

            if 'ssh_agent' in uri_query:
                self.__ssh_agent = uri_query.single_parameter('ssh_agent', bool)

            if 'private_key_file' in uri_query:
                self.__private_key_file = uri_query.single_parameter('private_key_file', str)

        if self.__uri.password is None and self.__private_key_file is None:
            raise ValueError('Password or private key file must be specified')

    def connect(self) -> None:
        """Synchronous implementation of the `.IOClientProto.connect` method
        """

        assert(self.__uri.hostname is not None)

        try:
            self.__ssh_client = paramiko.SSHClient()

            if self.__host_key_auto_add:
                self.__ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            self.__ssh_client.connect(
                hostname=self.__uri.hostname,
                port=(self.__uri.port if self.__uri.port is not None else 22),
                username=self.__uri.username,
                password=self.__uri.password,
                key_filename=self.__private_key_file,
                allow_agent=self.__ssh_agent
            )

            self.__sftp_client = self.__ssh_client.open_sftp()

            if self.__uri.path is not None:
                self.change_directory(self.__uri.path)
        except paramiko.ssh_exception.SSHException as e:
            self.disconnect()  # resets internal vars
            raise ConnectionError('Unable to connect') from e

    def disconnect(self) -> None:
        """Synchronous implementation of the `.IOClientProto.disconnect` method
        """
        if self.__sftp_client:
            self.__sftp_client.close()
            self.__sftp_client = None

        if self.__ssh_client:
            self.__ssh_client.close()
            self.__ssh_client = None

    def change_directory(self, path: str) -> str:
        """Synchronous implementation of the `.IOClientProto.change_directory` method
        """
        assert(self.__sftp_client is not None)

        if path in ('.', ''):
            return path_to_str(self.__vd_client.session_path())

        posix_path = pathlib.PosixPath(path)
        if not posix_path.is_absolute():
            posix_path = self.__vd_client.session_path() / posix_path

        try:
            self.__sftp_client.chdir(path_to_str(posix_path))
        finally:
            self.__sftp_client.chdir('/')

        normalized_path = pathlib.PosixPath(path_to_str(posix_path))
        return path_to_str(self.__vd_client.session_path(normalized_path))

    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def make_directory(self, directory_name: str) -> None:
        """Synchronous implementation of the `.IOClientProto.make_directory` method
        """
        assert(self.__sftp_client is not None)
        new_dir_path = self.__vd_client.entry_path(directory_name)
        self.__sftp_client.mkdir(str(new_dir_path))

    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def remove_directory(self, directory_name: str) -> None:
        """Synchronous implementation of the `.IOClientProto.remove_directory` method
        """
        assert(self.__sftp_client is not None)
        rm_dir_path = self.__vd_client.entry_path(directory_name)
        self.__sftp_client.rmdir(str(rm_dir_path))

    def list_directory(self) -> typing.Tuple[str, ...]:
        """Synchronous implementation of the `.IOClientProto.list_directory` method
        """
        assert(self.__sftp_client is not None)

        dir_to_list = path_to_str(self.__vd_client.session_path())
        return tuple(self.__sftp_client.listdir(dir_to_list))

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    def upload_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        """Synchronous implementation of the `.IOClientProto.upload_file` method
        """
        assert(self.__sftp_client is not None)

        local_file_obj.seek(0)
        data_length = local_file_obj.seek(0, os.SEEK_END)
        local_file_obj.seek(0)

        new_file_path = self.__vd_client.entry_path(remote_file_name)
        self.__sftp_client.putfo(local_file_obj, str(new_file_path), file_size=data_length)

    @verify_value(file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def remove_file(self, file_name: str) -> None:
        """Synchronous implementation of the `.IOClientProto.remove_file` method
        """
        assert(self.__sftp_client is not None)
        rm_file_path = self.__vd_client.entry_path(file_name)
        self.__sftp_client.remove(str(rm_file_path))

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    def receive_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        """Synchronous implementation of the `.IOClientProto.receive_file` method
        """
        assert(self.__sftp_client is not None)

        local_file_obj.seek(0)
        local_file_obj.truncate()
        file_path = self.__vd_client.entry_path(remote_file_name)

        self.__sftp_client.getfo(str(file_path), local_file_obj)

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def file_size(self, remote_file_name: str) -> int:
        """Synchronous implementation of the `.IOClientProto.file_size` method
        """
        assert(self.__sftp_client is not None)

        file_path = self.__vd_client.entry_path(remote_file_name)
        stat = self.__sftp_client.stat(str(file_path))

        if stat.st_size is not None:
            return stat.st_size

        raise IOError('File size not available')


@register_api(__default_io_clients_registry__, "sftp")
class SFTPClient(VirtualDirectoryClient):
    """This is an asynced SFTP client.
    """

    def __init__(self, uri: URI) -> None:
        """Create a new SFTP client.

        :param uri: URI to connect with
        """
        VirtualDirectoryClient.__init__(self, uri)
        self.__i12n = _SFTPSyncImplementation(uri, self)

        self._wrap_capability(self.__i12n, 'connect')
        self._wrap_capability(self.__i12n, 'disconnect')
        self._wrap_capability(self.__i12n, 'change_directory')
        self._wrap_capability(self.__i12n, 'list_directory')
        self._wrap_capability(self.__i12n, 'make_directory')
        self._wrap_capability(self.__i12n, 'remove_directory')
        self._wrap_capability(self.__i12n, 'upload_file')
        self._wrap_capability(self.__i12n, 'remove_file')
        self._wrap_capability(self.__i12n, 'receive_file')
        self._wrap_capability(self.__i12n, 'file_size')

    def current_directory(self) -> str:
        """The :meth:`.IOClientProto.current_directory` method implementation."""
        return str(self.session_path())



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/threaded_task.py
#
# Copyright (C) 2016-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import threading
import traceback
import typing

from pyknic.lib.log import Logger
from pyknic.lib.capability import iscapable
from pyknic.lib.signals.extra import AsyncWatchDog
from pyknic.lib.signals.proto import Signal
from pyknic.lib.tasks.proto import TaskProto, TaskResult, TaskStartError
from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.thread import CriticalResource


class ThreadedTask(TaskProto, CriticalResource):
    """ This class helps to run a task in a separate thread
    """

    thread_ready = Signal(TaskProto)  # signal is emitted when a thread is ready to join
    thread_joined = Signal(TaskProto)  # signal is emitted when a thread joined

    def __init__(
        self,
        task: TaskProto,
        cr_timeout: typing.Union[int, float, None] = None,
        thread_name: typing.Optional[str] = None
    ):
        """ Create a task that will start a thread

        :param task: a task to run in a thread
        :param cr_timeout: timeout for exclusive access to methods (same as timeout
        in the :meth:`.CriticalResource.__init__` method)
        :param thread_name: name of a thread to create
        """
        TaskProto.__init__(self)
        CriticalResource.__init__(self, cr_timeout)
        self.__task = task
        self.__thread = None
        self.__thread_name = thread_name

        if iscapable(self.__task, TaskProto.stop):
            self.append_capability(TaskProto.stop, self.__stop)

        if iscapable(self.__task, TaskProto.terminate):
            self.append_capability(TaskProto.terminate, self.__terminate)

    @CriticalResource.critical_section
    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        if self.__thread is not None:
            raise TaskStartError('A task is started already')

        self.__thread = threading.Thread(target=self.__threaded_function, name=self.__thread_name)
        self.__thread.start()

    def __threaded_function(self) -> None:
        """ Function that will be executed by a thread
        """
        self.emit(self.task_started)
        try:
            self.__task.start()  # result is always None
        except Exception as e:
            Logger.error(f'The "{self.__task_id()}" task failed with an error: {e}\n{traceback.format_exc()}')
            self.emit(self.thread_ready, self.__task)
            self.emit(self.task_completed, TaskResult(exception=e))
            return

        self.emit(self.thread_ready, self.__task)
        self.emit(self.task_completed, TaskResult())

    def __task_id(self) -> str:
        """ Return id of a task that is executed
        """
        return self.__task.task_name() if self.__task.task_name() else str(self.__task)  # type: ignore[return-value]

    @CriticalResource.critical_section
    def __stop(self) -> None:
        """ The :meth:`.TaskProto.stop` method implementation. Will be used if an original task has a stop capability
        """
        self.__task.stop()

    @CriticalResource.critical_section
    def __terminate(self) -> None:
        """ The :meth:`.TaskProto.terminate` method implementation. Will be used if an original task has a termination
        capability
        """
        self.__task.terminate()

    @CriticalResource.critical_section
    def join(self) -> bool:
        """ Join started thread. Every thread must be joined at least once

        :return: True if thread was joined, False otherwise
        """
        is_alive = False
        if self.__thread is not None:
            self.__thread.join(0)
            is_alive = self.__thread.is_alive()
            if not is_alive:
                self.__thread = None
                self.emit(ThreadedTask.thread_joined, self.__task)
        return not is_alive

    def wait(self, timeout: typing.Union[int, float, None] = None) -> None:
        """ Do not call a threaded task to stop, but wait till it stops

        :param timeout: a timeout with which this method should be called. If value is None then this function
        will wait forever, if value is negative or zero, then this function will poll current state,
        otherwise -- number of seconds to wait for

        :note: Use this method without timeout as a last resort only. Since it blocks other methods of this class
        """
        with self.critical_context():

            if self.__thread is not None:
                if timeout is None:
                    self.__thread.join()
                else:
                    self.__thread.join(timeout=timeout)

        self.join()

    async def start_async(self) -> None:
        """Start the thread and wait to finish in asyncio-way
        """

        watchdog = AsyncWatchDog(asyncio.get_event_loop(), self, ThreadedTask.thread_ready)
        self.start()
        await watchdog.wait()
        self.wait()  # force thread to join since callback may be received earlier
        self.join()

    @staticmethod
    def plain_task(
        fn: typing.Callable[[], typing.Any], cr_timeout: typing.Union[int, float, None] = None
    ) -> 'ThreadedTask':
        """ Return ThreadedTask that is based by the simple callable
        """
        return ThreadedTask(PlainTask(fn), cr_timeout)

    def __del__(self) -> None:
        """ Finalize this object and check that thread was joined
        """
        # TODO: create a test for this (straightforward test implementation didn't work)
        if self.__thread:
            raise RuntimeError(f"A thread wasn't awaited (the task -- {self.__task_id()})")



# -*- coding: utf-8 -*-
import logging

import pytest
import typing
import uuid

from pyknic.tasks.log import LogTask
from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.log import Logger


@pytest.fixture
def reset_logger() -> typing.Generator[None, None, None]:
    yield None
    current_handlers = Logger.handlers.copy()
    Logger.setLevel(logging.NOTSET)
    for h in current_handlers:
        Logger.removeHandler(h)


class TestLogTask:

    def test(self, monkeypatch: pytest.MonkeyPatch, empty_datalog: DatalogProto, reset_logger: None) -> None:
        assert(LogTask.dependencies() is None)
        task = LogTask.create(empty_datalog, 'sample-api-id', uuid.uuid4())
        assert(task.task_name() is not None)

        with pytest.raises(ValueError):
            task.start()

        monkeypatch.setenv(LogTask.__env_var_name__, "FOO")

        with pytest.raises(ValueError):
            task.start()

        monkeypatch.setenv(LogTask.__env_var_name__, "DEBUG")
        task.start()



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod

import pytest

from pyknic.lib.registry import APIRegistry, DuplicateAPIIdError
from pyknic.lib.i12n import register_implementation


def test() -> None:
    registry = APIRegistry()

    class Proto(metaclass=ABCMeta):

        @abstractmethod
        def foo(self) -> None:
            raise NotImplementedError('!')

    @register_implementation(registry, Proto)
    class Implementation(Proto):
        def foo(self) -> None:
            pass

    assert(registry[Proto] is Implementation)

    with pytest.raises(DuplicateAPIIdError):
        @register_implementation(registry, Proto)
        class AlternateImplementation(Proto):
            def foo(self) -> None:
                pass

    with pytest.raises(NotImplementedError):
        @register_implementation(registry, Proto)
        class NotAImplementation:
            def foo(self) -> None:
                pass



# -*- coding: utf-8 -*-
# pyknic/lib/gettext.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import functools
import gettext
import pathlib


class GetTextWrapper:
    """ This class wraps base input parameters for gettext calls. It simplifies and shortens gettext usage
    """

    def __init__(self, translations_path: pathlib.Path, fallback_lang: str | None = None):
        """ Create a wrapper with translations in a specified path

        :param translations_path: directory with language subdirectories
        :param fallback_lang: language to be used if a requested one isn't known (by default the "en_US" is used)
        """
        self.__loc_path = translations_path
        self.__fallback_lang = fallback_lang if fallback_lang else 'en_US'

    def lang(self, lang_name: str | None = None) -> gettext.GNUTranslations | gettext.NullTranslations:
        """ Return translation for a specified language

        :param lang_name: language to return when defined, otherwise a fallback language will be requested
        """
        if lang_name:
            return self.translation(lang_name, str(self.__loc_path))
        return self.translation(self.__fallback_lang, str(self.__loc_path))

    def __call__(self, lang_name: str | None = None) -> gettext.GNUTranslations | gettext.NullTranslations:
        """ Synonym for the :meth:`.GetTextWrapper.lang` method call
        """
        return self.lang(lang_name)

    @staticmethod
    @functools.cache
    def translation(lang_name: str, localedir: str) -> gettext.GNUTranslations | gettext.NullTranslations:
        """ Return a translation for a specified language and directory

        :param lang_name: language to use
        :param localedir: directory with language subdirectories
        """
        return gettext.translation("messages", localedir=localedir, languages=[lang_name], fallback=True)



# -*- coding: utf-8 -*-
# pyknic/lib/crypto/hash.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import abstractmethod

from cryptography.hazmat.primitives import hashes as c10y_hashes
from cryptography.hazmat.primitives.hashes import HashAlgorithm as C10yHashAlgorithm

from pyknic.lib.io import IOGenerator, IOProducer
from pyknic.lib.registry import APIRegistry, register_api
from pyknic.lib.crypto.proto import HasherProto

__default_hashers_registry__ = APIRegistry()


class CryptographyHasherAdapter(HasherProto):
    """This is an adapter for cryptography hashers.
    """

    def __init__(self) -> None:
        """Create a new hasher."""
        HasherProto.__init__(self)
        self.__hasher = self.__create_hasher()
        self.__digest: typing.Optional[bytes] = None

    @abstractmethod
    def _hasher_name(self) -> str:
        """Return class name from the cryptography library that implement required hash algorithm."""
        raise NotImplementedError('This method is abstract')

    def _hasher_args(self) -> typing.Tuple[typing.Any, ...]:
        """Return args required for hasher object creation
        """
        return tuple()

    def __create_hasher(self) -> c10y_hashes.Hash:
        """Create a new object from the cryptography library."""
        hash_algo = self.c10y_algorithm()
        if hash_algo is None:
            raise RuntimeError('No suitable algorithm available')

        return c10y_hashes.Hash(hash_algo)

    def update(self, source: IOProducer) -> IOGenerator:
        """The :meth:`.HasherProto.update` method implementation."""
        self.__hasher = self.__create_hasher()

        for chunk in source:
            self.__hasher.update(chunk)
            yield chunk

        self.__digest = self.__hasher.finalize()

    def digest(self) -> bytes:
        """The :meth:`.HasherProto.digest` method implementation."""
        if self.__digest is None:
            raise ValueError('Digest is not available. Please call the "update" method first.')
        return self.__digest

    def c10y_algorithm(self) -> typing.Optional[C10yHashAlgorithm]:
        """The :meth:`.HasherProto.c10y_algorithm` method implementation."""
        hash_algo = getattr(c10y_hashes, self._hasher_name())
        hash_args = self._hasher_args()

        return hash_algo(*hash_args)  # type: ignore[no-any-return]


@register_api(__default_hashers_registry__, "blake2b_64")
class BLAKE2b_64Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'BLAKE2b'

    def _hasher_args(self) -> typing.Tuple[typing.Any, ...]:
        """The :meth:`.CryptographyHasherAdapter._hasher_args` method implementation."""
        return (64, )


@register_api(__default_hashers_registry__, "blake2s_32")
class BLAKE2s_32Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'BLAKE2s'

    def _hasher_args(self) -> typing.Tuple[typing.Any, ...]:
        """The :meth:`.CryptographyHasherAdapter._hasher_args` method implementation."""
        return (32, )


@register_api(__default_hashers_registry__, "md5")
class MD5Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'MD5'


@register_api(__default_hashers_registry__, "sha1")
class SHA1Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA1'


@register_api(__default_hashers_registry__, "sha224")
class SHA224Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA224'


@register_api(__default_hashers_registry__, "sha256")
class SHA256Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA256'


@register_api(__default_hashers_registry__, "sha384")
class SHA384Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA384'


@register_api(__default_hashers_registry__, "sha512")
class SHA512Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA512'


@register_api(__default_hashers_registry__, "sha512_224")
class SHA512_224Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA512_224'


@register_api(__default_hashers_registry__, "sha512_256")
class SHA512_256Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA512_256'


@register_api(__default_hashers_registry__, "sha3_224")
class SHA3_224Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA3_224'


@register_api(__default_hashers_registry__, "sha3_256")
class SHA3_256Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA3_256'


@register_api(__default_hashers_registry__, "sha3_384")
class SHA3_384Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA3_384'


@register_api(__default_hashers_registry__, "sha3_512")
class SHA3_512Hasher(CryptographyHasherAdapter):

    def _hasher_name(self) -> str:
        """The :meth:`.CryptographyHasherAdapter._hasher_name` method implementation."""
        return 'SHA3_512'



# -*- coding: utf-8 -*-

import threading
import typing
import pytest

from pyknic.lib.tasks.proto import TaskProto, ScheduleSourceProto
from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.datalog.datalog import Datalog
from pyknic.lib.tasks.scheduler.record import ScheduleRecord


class SampleTasks:

    class DummyTask(TaskProto):

        def start(self) -> None:
            pass

    class LongRunningTask(TaskProto):

        def __init__(self, stop_method: bool = True, terminate_method: bool = True):
            TaskProto.__init__(self)
            self.__event = threading.Event()

            if stop_method:
                self.append_capability(TaskProto.stop, self.__stop_func)

            if terminate_method:
                self.append_capability(TaskProto.terminate, self.__stop_func)

        def start(self) -> None:
            self.__event.clear()
            self.__event.wait()

        def __stop_func(self) -> None:
            self.__event.set()

    class PlainRecord(ScheduleRecord):

        def __init__(self, task: TaskProto, **kwargs: typing.Any) -> None:
            ScheduleRecord.__init__(self, task, SampleTasks.DummySource(), **kwargs)

    class DummySource(ScheduleSourceProto):
        pass


@pytest.fixture
def sample_tasks() -> type:
    return SampleTasks


@pytest.fixture
def empty_datalog() -> DatalogProto:
    return Datalog()



# -*- coding: utf-8 -*-
# pyknic/lib/datalog/proto.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import abstractmethod

from pyknic.lib.signals.proto import Signal
from pyknic.lib.signals.source import SignalSource


class DatalogProto(SignalSource):
    """ This class represent sequence of data, where new data may be updated to the tail only
    """

    new_entry = Signal(object)

    @abstractmethod
    def append(self, record: typing.Any) -> None:
        """ Append new data to the end of sequence

        :param record: data to update
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def iterate(self, reverse: bool = False) -> typing.Generator[typing.Any, None, None]:
        """ Return iterator that yields over saved data from the oldest one to a newer one

        :param reverse: whether to return from the newer record to the oldest one (in case of True) and
        return in default order otherwise
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def truncate(self, min_length: int) -> None:
        """ Remove old records and keep at least N records

        :param min_length: number of records in sequence to keep. Not less than this number of records will be kept
        """
        raise NotImplementedError('This method is abstract')

    def find(
        self, filter_fn: typing.Callable[[typing.Any], bool], reverse: bool = False
    ) -> typing.Optional[typing.Any]:
        """ Try to find the closest entry that is suitable to the filter function

        :param filter_fn: a function that returns True for suitable log entry
        :param reverse: if False return the oldest record and return the newest record otherwise
        """
        raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# pyknic/tasks/gettext.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.tasks.scheduler.chain_source import ChainedTask, __default_chained_tasks_registry__
from pyknic.lib.registry import register_api
from pyknic.lib.log import Logger
from pyknic.lib.gettext import GetTextWrapper
from pyknic.path import root_path


@register_api(__default_chained_tasks_registry__, "gettext_task")
class GetTextInitTask(ChainedTask):
    """ This task creates a pre-defined localization object (GetTextWrapper)
    """

    __translations_loc__ = root_path / 'locales'  # where localization files are stored

    def start(self) -> None:
        """ The :meth:`.ChainedTask.start` method implementation
        """
        self.wait_for('log_task')
        Logger.info('Setting up a localization structure')

        translations = GetTextWrapper(self.__translations_loc__.absolute())
        self.save_result(translations)

    def task_name(self) -> str:
        """ Return this task's name
        """
        return "gettext-init"

    @classmethod
    def dependencies(cls) -> typing.Optional[typing.Set[str]]:
        """ The :meth:`.ChainedTask.dependencies` method implementation
        """
        return {"log_task"}



# -*- coding: utf-8 -*-
# pyknic/lib/io/compression.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Think of https://pypi.org/project/lz4/

import bz2
import gzip
import io
import lzma
import os
import typing

from abc import ABCMeta, abstractmethod

from pyknic.lib.registry import APIRegistry, register_api
from pyknic.lib.io import __default_block_size__, IOGenerator, IOProducer


__default_io_compressors_registry__ = APIRegistry()


class CompressorProto(metaclass=ABCMeta):
    """This is the base class for all compressors.
    """

    @abstractmethod
    def compress(self, source: IOProducer) -> IOGenerator:
        """Compress data and yield compressed chunks

        :param source: a data to compress
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def decompress(self, source: IOProducer) -> IOGenerator:
        """Decompress data and yield uncompressed chunks

        :param source: a compressed data
        """
        raise NotImplementedError('This method is abstract')


class NativeCompressor(CompressorProto):
    """This is an adapter for CPython compressors.
    """

    @abstractmethod
    def _compressor(self, io_buffer: typing.BinaryIO, mode: str) -> io.BufferedRWPair:
        """Return object that has common compressor methods that every CPython compressor has.

        :param io_buffer: an internal buffer that compressor works with
        :param mode: a mode with which a compressor should be opened (like 'rb' or 'wb')
        """
        raise NotImplementedError('This method is abstract')

    def compress(self, source: IOProducer) -> IOGenerator:
        """The :meth:`.CompressorProto.compress` method implementation."""
        compress_buffer = io.BytesIO()
        compressor = self._compressor(compress_buffer, 'wb')

        for data in source:
            compressor.write(data)

            comp_data = compress_buffer.getvalue()
            yield comp_data

            compress_buffer.truncate(0)
            compress_buffer.seek(0, os.SEEK_SET)

        compressor.flush()
        compressor.close()

        yield compress_buffer.getvalue()

    def decompress(self, source: IOProducer) -> IOGenerator:
        """The :meth:`.CompressorProto.decompress` method implementation."""
        compress_buffer = io.BytesIO()
        compressor = self._compressor(compress_buffer, 'rb')

        for data in source:
            compress_buffer.write(data)
            compress_buffer.seek(0, os.SEEK_SET)

            try:
                while True:
                    pos = compressor.tell()
                    chunk = compressor.peek(__default_block_size__)
                    if not chunk:
                        break
                    yield chunk
                    compressor.seek(pos + len(chunk), os.SEEK_SET)

            except EOFError:
                pass

            compress_buffer.truncate(0)
            compress_buffer.seek(0, os.SEEK_SET)

        yield compressor.read()


@register_api(__default_io_compressors_registry__, "gzip")
class GZipCompressor(NativeCompressor):
    """GZip implementation."""

    def _compressor(self, io_buffer: typing.BinaryIO, mode: str) -> io.BufferedRWPair:
        """The :meth:`.NativeCompressor._compressor` method implementation."""
        return gzip.GzipFile(fileobj=io_buffer, mode=mode)  # type: ignore[return-value]


@register_api(__default_io_compressors_registry__, "bzip2")
class BZip2Compressor(NativeCompressor):
    """BZip2 implementation."""

    def _compressor(self, io_buffer: typing.BinaryIO, mode: str) -> io.BufferedRWPair:
        """The :meth:`.NativeCompressor._compressor` method implementation."""
        return bz2.BZ2File(io_buffer, mode=mode)  # type: ignore[no-any-return, call-overload]


@register_api(__default_io_compressors_registry__, "lzma")
class LZMACompressor(NativeCompressor):
    """LZMA implementation."""

    def _compressor(self, io_buffer: typing.BinaryIO, mode: str) -> io.BufferedRWPair:
        """The :meth:`.NativeCompressor._compressor` method implementation."""
        return lzma.LZMAFile(io_buffer, mode=mode)  # type: ignore[return-value]



# -*- coding: utf-8 -*-

import typing
import pytest

from pyknic.lib.signals.proto import Signal, UnknownSignalException, SignalSourceProto, SignalCallbackProto
from pyknic.lib.signals.proto import SignalProxyProto


def test_exceptions() -> None:
    assert(issubclass(UnknownSignalException, Exception) is True)


def test_abstract() -> None:
    pytest.raises(TypeError, SignalSourceProto)
    pytest.raises(NotImplementedError, SignalSourceProto.emit, None, Signal())
    pytest.raises(NotImplementedError, SignalSourceProto.callback, None, Signal(), lambda: None)
    pytest.raises(NotImplementedError, SignalSourceProto.remove_callback, None, Signal(), lambda: None)

    pytest.raises(TypeError, SignalCallbackProto)
    pytest.raises(NotImplementedError, SignalCallbackProto.__call__, None, None, Signal())

    pytest.raises(TypeError, SignalProxyProto)
    pytest.raises(NotImplementedError, SignalProxyProto.proxy, None, None)
    pytest.raises(NotImplementedError, SignalProxyProto.discard_proxy, None, None)


class TestSignal:

    def test(self) -> None:
        signal = Signal()
        _ = {signal: 1}  # check that signals may be used as dict keys
        assert(signal != Signal())
        assert(signal == signal)

        assert(repr(signal) == object.__repr__(signal))

        signal.__pyknic_signal_name__ = 'foo'
        assert (repr(signal) == 'foo')

        signal.__pyknic_signal_name__ = 'bar'
        assert (repr(signal) == 'bar')

    @pytest.mark.parametrize('signal_args,signal_value,exc', [
        (tuple(), 1, None),
        (tuple(), 'foo', None),
        ((int, ), 1, None),
        ((int, ), 'foo', TypeError),
        ((int, lambda x: x < 10), 1, None),
        ((int, lambda x: x > 10), 1, ValueError),
    ])
    def test_check(
        self, signal_args: typing.Tuple[typing.Any], signal_value: typing.Any, exc: typing.Optional[typing.Any]
    ) -> None:
        signal = Signal(*signal_args)
        if exc:
            with pytest.raises(exc):
                signal.check_value(signal_value)
        else:
            signal.check_value(signal_value)  # raises nothing



# -*- coding: utf-8 -*-
# pyknic/lib/io/tar.py
#
# Copyright (C) 2025-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import dataclasses
import datetime
import grp
import io
import os
import pathlib
import pwd
import tarfile
import typing

from abc import ABCMeta, abstractmethod

from pyknic.lib.verify import verify_value
from pyknic.lib.io import IOGenerator, IOProducer
from pyknic.lib.io.aio_wrapper import IOThrottler, cag
from pyknic.lib.io.aligner import Aligner


MAX_DYNAMIC_FILE_SIZE = (2 ** 10) ** 7  # This is a maximum number of bytes (1 zebibyte) that may be saved "dynamically"
# Dynamically means that a size of data is unknown at archiving startup.
# Units:
#   kibi  Ki 2^10
#   mebi  Mi 2^20
#   gibi  Gi 2^30
#   tebi  Ti 2^40
#   pebi  Pi 2^50
#   exbi  Ei 2^60
#   zebi  Zi 2^70
#   yobi  Yi 2^80
#   robi  Ri 2^90
#   quebi Qi 2^100
#
# Note. The maximum volume size for EXT4 is 1 exbibyte


class StaticTarEntryProto(metaclass=ABCMeta):
    """This class describes tar entry which size is known before start
    """

    @abstractmethod
    def entry(self) -> IOGenerator:
        """Yield bytes that describe tar header and data.
        """
        raise NotImplementedError('This method is abstract')


class DynamicTarEntryProto(metaclass=ABCMeta):
    """This class describes tar entry which size is unknown

    :note: please note the MAX_DYNAMIC_FILE_SIZE
    """

    @abstractmethod
    def entry(self, file_obj: typing.IO[bytes]) -> IOGenerator:
        """Yield bytes that describe tar header and data.

        :param file_obj: file object to which data is written. Header and content must be yielded, but this object
        may be used for patching header only
        """
        raise NotImplementedError('This method is abstract')


class _TarPaddingGenerator:
    """This class calculates required padding and write it
    """

    @verify_value(aligned_to=lambda x: x > 0)
    def __init__(self, aligned_to: int, extra_padding: bool = False):
        """Create new padding helper

        :param aligned_to: number of bytes required for padding
        :param extra_padding: whether extra padding block should be appended at the end
        """
        self.__aligned_to = aligned_to if aligned_to else tarfile.BLOCKSIZE
        self.__extra_padding = extra_padding
        self.__counter = 0

    def _write(self, data: IOProducer) -> IOGenerator:
        """This processor writes data and required padding

        :param data: data to write
        """
        for block in data:
            self.__counter += len(block)
            yield block

        _, extra_bytes = divmod(self.__counter, self.__aligned_to)
        delta = (self.__aligned_to - extra_bytes) if extra_bytes else 0
        if self.__extra_padding:
            delta += self.__aligned_to
        yield (tarfile.NUL * delta) if delta else b''


class _BaseTarInnerFileGenerator(_TarPaddingGenerator):
    """This class helps to save a custom entry inside a tar archive.
    """

    def __init__(self) -> None:
        """Create a new file writer
        """
        _TarPaddingGenerator.__init__(self, tarfile.BLOCKSIZE)

    # noinspection PyMethodMayBeStatic
    def __pax_headers_padding(self, tar_info: tarfile.TarInfo, expected_size: int) -> tarfile.TarInfo:
        """Fill comment section in PAX header in order to comply limitations

        :param tar_info: original header to update
        :param expected_size: expected size of tar header
        """

        if expected_size % tarfile.BLOCKSIZE != 0:
            raise ValueError('Expected tar header is not aligned with the blocksize')

        if len(tar_info.tobuf()) > expected_size:
            raise ValueError('Original tar header is more than expected size')

        comment_chunk = 'pyknic-' + ('-c' * 60) + ';'  # 128 bytes
        tar_info.pax_headers['comment'] = comment_chunk  # type: ignore[index]
        while len(tar_info.tobuf()) < expected_size:
            tar_info.pax_headers['comment'] += comment_chunk  # type: ignore[index]

        if len(tar_info.tobuf()) != expected_size:
            raise ValueError('Unable to align tar header')

        return tar_info

    def _custom_tar_info(
        self,
        filename: str,
        /,
        size: typing.Optional[int] = None,
        mtime: typing.Optional[int] = None,
        mode: typing.Optional[int] = None,
        uid: typing.Optional[int] = None,
        gid: typing.Optional[int] = None,
        uname: typing.Optional[str] = None,
        gname: typing.Optional[str] = None,
        expected_header_size: typing.Optional[int] = None,
    ) -> tarfile.TarInfo:
        """Generate meta-info for a single file inside archive.

        :param filename: filename of an inner file.
        :param size: size of file (zero by default).
        :param mtime: file modification time (is now by default).
        :param mode: file mode (0440 by default).
        :param uid: file UID (current process user is used by default).
        :param gid: file GID (current process group is used by default).
        :param uname: username related to uid.
        :param gname: group name related to gid.
        :param expected_header_size: if specified then it is number of bytes that tar header should have.
        (result may have extra comment message in PAX header in order to achieve expectations)
        """
        if filename.startswith('/'):
            raise ValueError('Tar entries should not have absolute path because of possible tar bomb')

        tar_info = tarfile.TarInfo(name=filename)
        if size is not None:
            tar_info.size = size
        tar_info.mtime = mtime if mtime is not None else int(datetime.datetime.now().timestamp())
        tar_info.mode = mode if mode is not None else int('440', base=8)
        tar_info.type = tarfile.REGTYPE
        tar_info.uid = uid if uid is not None else os.getuid()
        tar_info.gid = gid if gid is not None else os.getgid()
        tar_info.uname = uname if uname is not None else pwd.getpwuid(tar_info.uid).pw_name
        tar_info.gname = gname if gname is not None else grp.getgrgid(tar_info.gid).gr_name

        if expected_header_size is not None and len(tar_info.tobuf()) != expected_header_size:
            tar_info = self.__pax_headers_padding(tar_info, expected_header_size)

        return tar_info

    # noinspection PyMethodMayBeStatic
    def _tar_info_by_file(self, filename: typing.Union[str, pathlib.Path]) -> tarfile.TarInfo:
        """Retrieve meta-info by a real file

        :param filename: file name which meta-info should be retrieved
        """
        # TODO: filename related to some directory?!
        tar_file = tarfile.TarFile('/dev/zero', mode='w')  # TODO: it seams bogus
        tar_info = tar_file.gettarinfo(filename)

        if tar_info.name.startswith('/'):
            raise ValueError('Tar entries should not have absolute path because of possible tar bomb')

        return tar_info


class TarInnerFileGenerator(_BaseTarInnerFileGenerator, StaticTarEntryProto):
    """This class helps to save a static file inside a tar archive.
    """

    def __init__(self, source: typing.Union[str, pathlib.Path]):
        """Create a new inner file generator

        :param source: a path to a file to archive
        """
        _BaseTarInnerFileGenerator.__init__(self)
        self.__source = source

    def entry(self) -> IOGenerator:
        """ :meth:`.StaticTarEntryProto.entry` implementation
        """

        def data_generator() -> IOGenerator:
            tar_info = self._tar_info_by_file(self.__source)
            yield tar_info.tobuf()

            with open(self.__source, 'rb') as f:
                # TODO: consider checking that file modification time hasn't changed!

                for chunk in IOThrottler.sync_reader(f):
                    yield chunk

        yield from self._write(data_generator())


class TarInnerGenerator(_BaseTarInnerFileGenerator, StaticTarEntryProto):
    """This class helps to save a static data (with known size) from generator inside a tar archive.
    """

    def __init__(
        self,
        source: IOProducer,
        data_size: int,
        filename: str,
        /,
        mtime: typing.Optional[int] = None,
        mode: typing.Optional[int] = None,
        uid: typing.Optional[int] = None,
        gid: typing.Optional[int] = None,
        uname: typing.Optional[str] = None,
        gname: typing.Optional[str] = None,
    ):
        """Create a new inner file generator

        :param source: a data to save
        :param filename: name of an inner file
        :param data_size: number of bytes that source has. It must be exact number of bytes (not more, not less)

        :param mtime: file modification time (is now by default)
        :param mode: file mode (0440 by default)
        :param uid: file UID (current process user is used by default)
        :param gid: file GID (current process group is used by default)
        :param uname: username related to uid
        :param gname: group name related to gid        """
        _BaseTarInnerFileGenerator.__init__(self)

        if filename.startswith('/'):
            raise ValueError('File path must be relative')

        self.__source = source
        self.__data_size = data_size
        self.__filename = filename
        self.__mtime = mtime
        self.__mode = mode
        self.__uid = uid
        self.__gid = gid
        self.__uname = uname
        self.__gname = gname

    def entry(self) -> IOGenerator:
        """ :meth:`.StaticTarEntryProto.entry` implementation
        """

        def data_generator() -> IOGenerator:
            tar_info = self._custom_tar_info(
                self.__filename,
                size=self.__data_size,
                mtime=self.__mtime,
                mode=self.__mode,
                uid=self.__uid,
                gid=self.__gid,
                uname=self.__uname,
                gname=self.__gname,
            )
            yield tar_info.tobuf()

            calculated_size = 0
            for data in self.__source:
                calculated_size += len(data)
                if calculated_size > self.__data_size:
                    raise ValueError(
                        'Original data has more bytes than expected. '
                        f'Expected -- {self.__data_size}, found -- {calculated_size}'
                    )

                yield data

            if calculated_size != self.__data_size:
                raise ValueError(
                    'Incorrect number of bytes in original data. '
                    f'Expected -- {self.__data_size}, found -- {calculated_size}'
                )

        yield from self._write(data_generator())


class TarInnerDynamicGenerator(_BaseTarInnerFileGenerator, DynamicTarEntryProto):
    """This class helps to save a dynamic data from generator (which size is unknown) inside a tar archive.
    """

    def __init__(
        self,
        source: IOProducer,
        filename: str,
        /,
        mtime: typing.Optional[int] = None,
        mode: typing.Optional[int] = None,
        uid: typing.Optional[int] = None,
        gid: typing.Optional[int] = None,
        uname: typing.Optional[str] = None,
        gname: typing.Optional[str] = None,
    ):
        """Create a new inner file generator

        :param source: a data to save
        :param filename: name of an inner file

        :param mtime: file modification time (is now by default)
        :param mode: file mode (0440 by default)
        :param uid: file UID (current process user is used by default)
        :param gid: file GID (current process group is used by default)
        :param uname: username related to uid
        :param gname: group name related to gid        """
        _BaseTarInnerFileGenerator.__init__(self)

        if filename.startswith('/'):
            raise ValueError('File path must be relative')

        self.__source = source
        self.__filename = filename
        self.__mtime = mtime
        self.__mode = mode
        self.__uid = uid
        self.__gid = gid
        self.__uname = uname
        self.__gname = gname

    @verify_value(file_obj=lambda x: x.seekable())
    def entry(self, file_obj: typing.IO[bytes]) -> IOGenerator:
        """ :meth:`.DynamicTarEntryProto.entry` implementation
        """

        def data_generator() -> IOGenerator:
            start_pos = file_obj.tell()

            tar_info = self._custom_tar_info(
                self.__filename,
                size=MAX_DYNAMIC_FILE_SIZE,
                mtime=self.__mtime,
                mode=self.__mode,
                uid=self.__uid,
                gid=self.__gid,
                uname=self.__uname,
                gname=self.__gname,
            )
            binary_tar_info = tar_info.tobuf()
            yield binary_tar_info

            data_size = 0
            for chunk in self.__source:
                data_size += len(chunk)

                if data_size > MAX_DYNAMIC_FILE_SIZE:
                    raise ValueError('Are you saving the universe?!')

                yield chunk

            end_pos = file_obj.tell()

            patched_tar_info = self._custom_tar_info(
                self.__filename,
                size=data_size,
                mtime=self.__mtime,
                mode=self.__mode,
                uid=self.__uid,
                gid=self.__gid,
                uname=self.__uname,
                gname=self.__gname,
                expected_header_size=len(binary_tar_info)
            )

            patched_binary_tar_info = patched_tar_info.tobuf()
            assert(len(binary_tar_info) == len(patched_binary_tar_info))

            file_obj.seek(start_pos, os.SEEK_SET)
            file_obj.write(patched_binary_tar_info)
            file_obj.seek(end_pos, os.SEEK_SET)

        yield from self._write(data_generator())


@dataclasses.dataclass
class _FilthyTarEntry:
    """This class helps to track tar entries while reading. Using it in other modules should be avoided. Better choose
    :class:`.DecentTarEntry`
    """

    source: IOGenerator           # source of the data
    binary_info: bytes            # original tar header (may be larger than tarfile.BLOCKSIZE when it is combined
    # from several extended entries)
    parsed_info: tarfile.TarInfo  # tar header

    def data(self) -> IOGenerator:
        """Read data related to this entry
        """
        counter = self.parsed_info.size

        while counter > 0:
            next_chunk = next(self.source)
            assert(len(next_chunk) == tarfile.BLOCKSIZE)
            counter -= tarfile.BLOCKSIZE
            yield next_chunk


class DecentTarEntry:
    """This class helps to read tar entry in a safe way.
    """

    def __init__(self, tar_info: tarfile.TarInfo, data_generator: IOGenerator) -> None:
        """Create new entry.

        :param tar_info: information about inner entry
        :param data_generator: content of inner entry that must be read before the request for the next entry.
        """
        self.__tar_info = tar_info
        self.__data_generator = data_generator
        self.__wasted = False

    def is_wasted(self) -> bool:
        """Return True if data fully read.
        """
        return self.__wasted

    def tar_info(self) -> tarfile.TarInfo:
        """Return information about inner entry
        """
        return self.__tar_info

    def data(self) -> IOGenerator:
        """Read data related to this entry
        """
        counter = self.__tar_info.size

        while counter > 0:
            next_chunk = next(self.__data_generator)
            assert(len(next_chunk) == tarfile.BLOCKSIZE)

            if counter >= tarfile.BLOCKSIZE:
                yield next_chunk
            else:
                yield next_chunk[:counter]
            counter -= tarfile.BLOCKSIZE

        self.__wasted = True


class _TarReader:
    """This class helps to read data from a tar archive. Since it scans through all the data it may be unacceptable
    slowly. Use it on you own risk
    """

    # noinspection PyMethodMayBeStatic
    def _is_regular_entry(self, entry: tarfile.TarInfo) -> bool:
        """Return True if entry is FS-representable entry like file, directory or some named socket.

        :param entry: entry to check
        """

        regular_entries = [
            tarfile.REGTYPE,
            tarfile.AREGTYPE,
            tarfile.LNKTYPE,
            tarfile.SYMTYPE,
            tarfile.CHRTYPE,
            tarfile.BLKTYPE,
            tarfile.DIRTYPE,
            tarfile.FIFOTYPE,

            # there is tarfile.CONTTYPE also, but what the fuck is this?!
        ]

        return entry.type in regular_entries

    # noinspection PyMethodMayBeStatic
    def _is_extended_head_entry(self, entry: tarfile.TarInfo) -> bool:
        """Return True if entry is a special one dedicated to overcome original TAR limits (like path limitations
        or size limitations)

        :param entry: entry to check
        """
        extended_head_types = [
            tarfile.GNUTYPE_LONGNAME,  # GNU format
            tarfile.GNUTYPE_LONGLINK,  # GNU format
            tarfile.GNUTYPE_SPARSE,    # GNU format
            tarfile.XHDTYPE            # PAX format

            # tarfile.XGLTYPE -- this is PAX format, but wtf?!
            # tarfile.SOLARIS_XHDTYPE -- some old shit
        ]

        return entry.type in extended_head_types

    # noinspection PyMethodMayBeStatic
    def _iterate_raw_entries(self, source: IOProducer) -> typing.Generator[_FilthyTarEntry, None, None]:
        """Iterate over all entries in source. Extended headers will be split and yielded independently
        """

        zero_block = b'\x00' * tarfile.BLOCKSIZE
        aligner = Aligner(tarfile.BLOCKSIZE, strict_mode=True)

        next_entry_size = 0
        aligner_generator = aligner.iterate_data(source)
        for chunk in aligner_generator:
            next_entry_size -= len(chunk)

            if chunk == zero_block:
                continue

            head = tarfile.TarInfo.frombuf(chunk, tarfile.ENCODING, 'strict')  # read the
            # https://docs.python.org/3/library/codecs.html#error-handlers
            yield _FilthyTarEntry(aligner_generator, chunk, head)

    def _iterate_extended_entries(self, source: IOProducer) -> typing.Generator[_FilthyTarEntry, None, None]:
        """Iterate over all entries in source. Extended headers will be combined and only "regular" entries will be
        yielded.
        """
        raw_entries_generator = self._iterate_raw_entries(source)
        for tar_entry in raw_entries_generator:
            tar_head = tar_entry.parsed_info

            if self._is_regular_entry(tar_head):
                yield tar_entry
                continue

            if self._is_extended_head_entry(tar_head):

                extended_head_data = tar_entry.binary_info
                for chunk in tar_entry.data():
                    extended_head_data += chunk

                next_entry = next(raw_entries_generator)
                while not self._is_regular_entry(next_entry.parsed_info):
                    if not self._is_extended_head_entry(next_entry.parsed_info):
                        raise ValueError(
                            f'Unknown tar entry spotted -- {str(tar_head.type)} (file -- {str(tar_head.name)})'
                        )

                    extended_head_data += next_entry.binary_info
                    for chunk in next_entry.data():
                        extended_head_data += chunk

                    next_entry = next(raw_entries_generator)

                extended_head_data += next_entry.binary_info

                combined_head = tarfile.TarFile(fileobj=io.BytesIO(extended_head_data)).next()
                assert(combined_head is not None)
                yield _FilthyTarEntry(next_entry.source, extended_head_data, combined_head)
                continue

            raise ValueError(f'Unknown tar entry spotted -- {str(tar_head.type)} (file -- {str(tar_head.name)})')

    def iterate_entries(self, data: IOProducer) -> typing.Generator[DecentTarEntry, None, None]:
        """Yield information about inner files
        """
        aligner = Aligner(tarfile.BLOCKSIZE, strict_mode=True)

        for filthy_entry in self._iterate_extended_entries(aligner.iterate_data(data)):
            previous_entry = DecentTarEntry(filthy_entry.parsed_info, filthy_entry.source)
            yield previous_entry

            if not previous_entry.is_wasted():
                raise RuntimeError('You must read data from previous entry before requesting a next one')


class TarArchive(_TarPaddingGenerator):
    """This class helps to save a single tar archive or helps to read it
    """

    def __init__(self) -> None:
        """Create a new tar archive writer
        """
        _TarPaddingGenerator.__init__(self, aligned_to=tarfile.RECORDSIZE, extra_padding=True)

    def static_archive(self, sources: typing.Iterable[StaticTarEntryProto]) -> IOGenerator:
        """Yield data that may be saved as a tar archive.

        :param sources: archive entries
        """

        def entries() -> IOGenerator:
            for s in sources:
                yield from s.entry()

        yield from self._write(entries())

    @verify_value(destination=lambda x: x.seekable())
    async def dynamic_archive(
        self,
        destination: typing.IO[bytes],
        sources: typing.Iterable[typing.Union[StaticTarEntryProto, DynamicTarEntryProto]],
        write_throttling: typing.Optional[int] = None
    ) -> None:
        """Save data as a tar archive.

        :param destination: destination file
        :param sources: archive entries
        :param write_throttling: whether to throttle at saving (number of bytes per second)
        """
        destination.seek(0)
        destination.truncate()

        def entries() -> IOGenerator:
            for s in sources:
                if isinstance(s, StaticTarEntryProto):
                    yield from s.entry()
                else:
                    assert(isinstance(s, DynamicTarEntryProto))
                    yield from s.entry(destination)

        await cag(IOThrottler.async_writer(self._write(entries()), destination, throttling=write_throttling))

    @staticmethod
    @verify_value(source=lambda x: x.seekable())
    def extract(source: typing.IO[bytes], filename: str) -> IOGenerator:
        """Extract data from an archive

        :param source: tar-data
        :param filename: name of a file to retrieve from archive
        """

        start_pos = source.tell()

        with tarfile.open(fileobj=source) as tar:
            tar_info = tar.getmember(filename)
            file_size = tar_info.size
            source.seek(start_pos + tar_info.offset_data, os.SEEK_SET)

            if file_size:  # not empty file
                yield from IOThrottler.sync_reader(source, read_size=file_size)

        source.seek(start_pos, os.SEEK_SET)

    @staticmethod
    def entries(source: IOProducer) -> typing.Generator[DecentTarEntry, None, None]:
        """Read data and yield information about inner files. Entries are returned one by one. Each entry must be
        read with the :meth:`.DecentTarEntry.data` method before the request for the next entry.

        :param source: tar data to read
        """
        yield from _TarReader().iterate_entries(source)

    @staticmethod
    def inner_descriptors(source: IOProducer) -> typing.Generator[tarfile.TarInfo, None, None]:
        """Yield information about inner files. Entries are returned one by one

        :param source: tar data to read
        """

        for i in _TarReader().iterate_entries(source):
            yield i.tar_info()
            for _ in i.data():
                pass



# -*- coding: utf-8 -*-

from pyknic.path import root_path
from pyknic.lib.gettext import GetTextWrapper


class TestGetTextWrapper:

    def test(self) -> None:
        wrapper = GetTextWrapper(root_path / 'locales')
        assert(wrapper.lang("ru").gettext("Choose a game") == "Выбирай игру")
        assert(wrapper("fr").gettext("Choose a game") == "Choose a game")  # no such translation
        assert(wrapper().gettext("Choose a game") == "Choose a game")  # no such translation



# -*- coding: utf-8 -*-

import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SampleTasks

from pyknic.lib.signals.proto import Signal, SignalSourceProto

from pyknic.lib.tasks.proto import ScheduleSourceProto
from pyknic.lib.tasks.scheduler.record import ScheduleRecord
from pyknic.lib.tasks.scheduler.plain_sources import InstantTaskSource


class TestInstantTaskSource:

    def test(self, sample_tasks: 'SampleTasks') -> None:
        callback_result = []

        def callback(src: SignalSourceProto, signal: Signal, value: ScheduleRecord) -> None:
            callback_result.append(value)

        source = InstantTaskSource()
        assert(isinstance(source, ScheduleSourceProto))

        record = ScheduleRecord(sample_tasks.DummyTask(), source)
        source.callback(source.task_scheduled, callback)
        source.schedule_record(record)
        assert(callback_result == [record])



# -*- coding: utf-8 -*-
# pyknic/lib/integrated_commands/logout.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import typing

import pydantic

from pyknic.lib.bellboy.app import register_bellboy_command, BellBoyCommandHandler
from pyknic.lib.bellboy.models import GeneralBellBoyCommandModel
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult, LobbyStrFeedbackResult


@register_bellboy_command()
class LogoutCommand(BellBoyCommandHandler):
    """ Forget known credential
    """

    @classmethod
    def command_name(cls) -> str:
        """The :meth:`.BellBoyCommandHandler.command_name` method implementation"""
        return "logout"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """The :meth:`.BellBoyCommandHandler.command_model` method implementation"""
        return GeneralBellBoyCommandModel

    async def exec(self) -> LobbyCommandResult:
        """The :meth:`.BellBoyInternalCommand.exec_from_cli` method implementation
        """
        assert(isinstance(self._args, GeneralBellBoyCommandModel))

        secret_backend = self.secret_backend(self._args.secret_backend)

        with contextlib.suppress(KeyError):
            secret_backend.pop_secret(self._args.lobby_url)
            return LobbyStrFeedbackResult(
                str_result=f'Login credential for the {self._args.lobby_url} site has been forgotten'
            )
        return LobbyStrFeedbackResult(
            str_result=f'There is no credential for the {self._args.lobby_url} site'
        )



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/scheduler.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import functools
import typing
import weakref

from pyknic.lib.capability import iscapable
from pyknic.lib.signals.proto import Signal, SignalSourceProto
from pyknic.lib.signals.proxy import QueueCallbackException
from pyknic.lib.signals.extra import BoundedCallback, SignalResender, CallbacksHolder
from pyknic.lib.verify import verify_value

from pyknic.lib.tasks.proto import SchedulerProto, ScheduleSourceProto, TaskProto, SchedulerFeedback, TaskStopError

from pyknic.lib.tasks.scheduler.scheduler_executor import SchedulerExecutor


class Scheduler(SchedulerProto, TaskProto):
    """ The :class:`.SchedulerProto` class implementation
    """

    def __init__(
        self,
        threads_number: typing.Optional[int] = None,
        executor_cr_timeout: typing.Union[int, float, None] = None,
        thread_cr_timeout: typing.Union[int, float, None] = None,
        task_timeout: typing.Union[int, float, None] = None
    ):
        """ Create a new scheduler with an internal thread executor

        :param threads_number: same as the "threads_number" argument in :meth:`.ThreadExecutor.__init__`
        :param executor_cr_timeout: same as the "executor_cr_timeout" argument in :meth:`.ThreadExecutor.__init__`
        :param thread_cr_timeout: same as the "thread_cr_timeout" argument in :meth:`.ThreadExecutor.__init__`
        :param task_timeout: same as the "task_timeout" argument in :meth:`SchedulerExecutor.await_tasks`
        """
        SchedulerProto.__init__(self)
        TaskProto.__init__(self)

        self.__executor = SchedulerExecutor(threads_number, executor_cr_timeout, thread_cr_timeout)
        self.__task_timeout = task_timeout
        self.__holder = CallbacksHolder()
        self.__sources: weakref.WeakSet[ScheduleSourceProto] = weakref.WeakSet()  # protects from double subscriptions
        # TODO: think of removing this "self.__sources" -- it will make code cleaner

        self.__task_scheduled_clbk = BoundedCallback(self.__task_scheduled)

        def link_signals(source_signal: Signal, target_signal: Signal) -> None:
            self.__executor.callback(
                source_signal,
                self.__holder.keep_callback(
                    SignalResender(self, target_signal=target_signal),
                    self
                )
            )

        link_signals(SchedulerExecutor.scheduled_task_dropped, SchedulerProto.scheduled_task_dropped)
        link_signals(SchedulerExecutor.scheduled_task_postponed, SchedulerProto.scheduled_task_postponed)
        link_signals(SchedulerExecutor.scheduled_task_expired, SchedulerProto.scheduled_task_expired)
        link_signals(SchedulerExecutor.scheduled_task_started, SchedulerProto.scheduled_task_started)
        link_signals(SchedulerExecutor.scheduled_task_completed, SchedulerProto.scheduled_task_completed)

    def __task_scheduled(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ A task (value) has been scheduled so this callback is processing it

        :param source: source (ScheduleSourceProto) that generates an event
        :param signal: the "ScheduleSourceProto.task_scheduled" signal
        :param value: aka ScheduleRecordProto -- a record to execute
        """
        self.emit(SchedulerProto.task_scheduled, value)
        self.__executor.submit(value, blocking=False)

    def __subscribe(self, schedule_source: ScheduleSourceProto) -> None:
        """ Subscribe to a source

        :param schedule_source: a source to subscribe
        """
        assert(self.__executor.queue_proxy().is_inside())

        if schedule_source in self.__sources:
            raise ValueError('Source is subscribed already')

        if iscapable(schedule_source, ScheduleSourceProto.scheduler_feedback):
            schedule_source.scheduler_feedback(self, SchedulerFeedback.source_subscribed)
        schedule_source.callback(ScheduleSourceProto.task_scheduled, self.__task_scheduled_clbk)
        self.__sources.add(schedule_source)

    @verify_value(timeout=lambda x: x is None or x > 0)
    def subscribe(self, schedule_source: ScheduleSourceProto) -> None:
        """ Request a source subscription

        :param schedule_source: a source to subscribe
        """
        assert(not self.__executor.queue_proxy().is_inside())  # this prevents queue_proxy from self-blocking
        try:
            self.__executor.queue_proxy().exec(functools.partial(self.__subscribe, schedule_source), blocking=True)
        except QueueCallbackException as e:
            raise e.__cause__  # type: ignore[misc]  # it's ok

    def __unsubscribe(self, schedule_source: ScheduleSourceProto) -> None:
        """ Unsubscribe a source

        :param schedule_source: a source to unsubscribe
        """
        assert(self.__executor.queue_proxy().is_inside())

        try:
            self.__sources.remove(schedule_source)
        except KeyError:
            raise ValueError('Unknown source is requested to unsubscribe')

        if iscapable(schedule_source, ScheduleSourceProto.scheduler_feedback):
            schedule_source.scheduler_feedback(self, SchedulerFeedback.source_unsubscribed)

        schedule_source.remove_callback(ScheduleSourceProto.task_scheduled, self.__task_scheduled_clbk)

    @verify_value(timeout=lambda x: x is None or x > 0)
    def unsubscribe(self, schedule_source: ScheduleSourceProto) -> None:
        """ Request a source "unsubscription"

        :param schedule_source: a source to unsubscribe
        """
        assert(not self.__executor.queue_proxy().is_inside())  # this prevents queue_proxy from self-blocking
        try:
            self.__executor.queue_proxy().exec(functools.partial(self.__unsubscribe, schedule_source), blocking=True)
        except QueueCallbackException as e:
            raise e.__cause__  # type: ignore[misc]  # it's ok

    def start(self) -> None:
        """ Start a schedule
        """
        self.__executor.queue_proxy().start()

    def __unsubscribe_all(self) -> None:
        """ Unsubscribe all sources as a part of stopping procedure
        """
        assert(self.__executor.queue_proxy().is_inside())

        for source in self.__sources.copy():
            if source:
                self.__unsubscribe(source)

    def stop(self) -> None:
        """ Stop this schedule
        """
        assert(not self.__executor.queue_proxy().is_inside())  # this prevents queue_proxy from self-blocking
        self.__executor.queue_proxy().exec(self.__unsubscribe_all, blocking=True)
        self.__executor.cancel_postponed_tasks()
        self.__executor.stop_running_tasks()
        self.__executor.await_tasks(self.__task_timeout)

        pending_tasks = self.__executor.pending_tasks()
        running_tasks = self.__executor.running_tasks()

        pts = len(pending_tasks)
        rts = len(running_tasks)
        if (pts + rts) > 0:
            raise TaskStopError(
                f'Scheduler still has running/pending task(s) (running -- {rts}; pending -- {pts})'  # noqa: E702
            )

        self.__executor.queue_proxy().stop()



# -*- coding: utf-8 -*-
# pyknic/lib/io/aligner.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.


from pyknic.lib.io import IOGenerator, IOProducer


class Aligner:
    """This class helps to chop IOGenerator into length-fixed blocks
    """

    def __init__(self, block_size: int, *, strict_mode: bool = False):
        """Create a new aligner

        :param block_size: size of block in bytes that should be yield
        :param strict_mode: whether to raise an exception when the last block was not aligned
        """
        self.__block_size = block_size
        self.__strict_mode = strict_mode

    def iterate_data(self, data: IOProducer) -> IOGenerator:
        """Chop data into blocks/chunks
        """

        data_cache = b''
        for i in data:
            data_cache += i

            while len(data_cache) > self.__block_size:
                yield data_cache[:self.__block_size]
                data_cache = data_cache[self.__block_size:]

        if len(data_cache) != self.__block_size and self.__strict_mode:
            raise RuntimeError(f'Data is not aligned within {self.__block_size} bytes')

        if data_cache:
            yield data_cache



# -*- coding: utf-8 -*-
# pyknic/lib/config.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

import yaml

from pyknic.lib.capability import CapabilitiesHolder, capability
from pyknic.lib.verify import verify_type, verify_value


class ConfigStorageProto(CapabilitiesHolder):
    """This class is used mostly by mypy to detect possible methods and results of methods. It is
    full of "capabilities" that are implemented in some classes and not in others.

    In a real implementation each object is similar to one of:
     - dict
     - list
     - str
     - bool
     - int
     - float
     - None
    and implement related methods only
    """

    @capability
    def getitem(self, item: typing.Union[str, int]) -> 'ConfigStorageProto':
        """Return value that is stored within a key, or get a value by an index (a real behaviour depends on
        implementation).

        :param item: key (index) which value should be retrieved
        """
        raise NotImplementedError('This method is abstract')

    def __getitem__(self, item: typing.Union[str, int]) -> 'ConfigStorageProto':
        """Shortcut for the :meth:`.ConfigStorageProto.getitem` capability."""
        return self.getitem(item)

    @capability
    def as_int(self) -> int:
        """If possible return integer value of this config. Suitable for a plain, non-aggregated entries."""
        raise NotImplementedError('This method is abstract')

    def __int__(self) -> int:
        """Shortcut for the :meth:`.ConfigStorageProto.as_int` capability."""
        return self.as_int()

    @capability
    def as_float(self) -> float:
        """If possible return floating point value of this config. Suitable for a plain, non-aggregated entries."""
        raise NotImplementedError('This method is abstract')

    def __float__(self) -> float:
        """Shortcut for the :meth:`.ConfigStorageProto.as_float` capability."""
        return self.as_float()

    @capability
    def as_bool(self) -> bool:
        """If possible return boolean value of this config. Suitable for a plain, non-aggregated entries."""
        raise NotImplementedError('This method is abstract')

    def __bool__(self) -> bool:
        """Shortcut for the :meth:`.ConfigStorageProto.as_bool` capability."""
        return self.as_bool()

    @capability
    def as_str(self) -> str:
        """If possible return string value of this config. Suitable for a plain, non-aggregated entries."""
        raise NotImplementedError('This method is abstract')

    def __str__(self) -> str:
        """Shortcut for the :meth:`.ConfigStorageProto.as_str` capability."""
        return self.as_str()

    @capability
    def is_none(self) -> bool:
        """Return True if this option is defined as None. Suitable for a plain, non-aggregated entries."""
        raise NotImplementedError('This method is abstract')

    @capability
    def properties(self) -> typing.Set[str]:
        """Return keys that this dict-a-like aggregator holds"""
        raise NotImplementedError('This method is abstract')

    @capability
    def has_property(self, name: str) -> bool:
        """Return True if this dict-a-like aggregator holds the specified key

        :param name: a key name to check
        """
        raise NotImplementedError('This method is abstract')

    @capability
    def property(self, name: str) -> 'ConfigStorageProto':
        """Return True if this dict-a-like aggregator holds the specified key

        :param name: a key name to check
        """
        raise NotImplementedError('This method is abstract')

    @capability
    def reset_properties(self) -> None:
        """Remove everything that this dict-a-like aggregator holds."""
        raise NotImplementedError('This method is abstract')

    @capability
    def merge_file(self, file_obj: typing.IO[str], property_name: typing.Optional[str] = None) -> None:
        """Union this dict-a-like configuration with another one.

        :param file_obj: file object to merge
        :param property_name: if defined then only this section will be imported from a file
        """
        raise NotImplementedError('This method is abstract')

    @capability
    def merge_config(self, config: 'ConfigStorageProto', property_name: typing.Optional[str] = None) -> None:
        """Union this dict-a-like configuration with another one.

        :param config: a config to read
        :param property_name: if defined then only this section will be imported from a config
        """
        raise NotImplementedError('This method is abstract')

    @capability
    def iterate_list(self) -> typing.Generator['ConfigStorageProto', None, None]:
        """Iterate over list-a-like configuration."""
        raise NotImplementedError('This method is abstract')


RawConfigPlainTypes: typing.TypeAlias = typing.Optional[typing.Union[int, float, str, bool]]

RawConfigComplexType: typing.TypeAlias = typing.Union[
    RawConfigPlainTypes,
    typing.List['RawConfigComplexType'],
    typing.Dict[str, typing.Union['RawConfigComplexType', '_ConfigImplementation']]
]


class _ConfigImplementation:
    """This is basic storage and basic app configuration implementation.

    It is private and helps to implement more complex logic.
    """

    def __init__(self, value: RawConfigComplexType | '_ConfigImplementation'):
        """Create a new configuration and initialize it with a value.

        :param value: initial value
        """

        @verify_type(init_value=(None, dict, list, int, float, str, bool, _ConfigImplementation))
        def init(init_value: RawConfigComplexType | '_ConfigImplementation') -> None:
            """Just decorated __init__ method that is called later."""
            self.__value: RawConfigComplexType = \
                init_value.__value if isinstance(init_value, _ConfigImplementation) else init_value

        init(value)

    def __ensure_type(self, *value_types: type) -> None:
        """Check that inside storage suites specified types.

        :param value_types: types that storage should be suited
        """
        assert(len(value_types) >= 1)

        if not isinstance(self.__value, value_types):
            raise TypeError('Invalid type for config entry')

    def is_none(self) -> bool:
        """Return that this configuration holds nothing ie None."""
        return self.__value is None

    def is_dict(self) -> bool:
        """Return that inside storage is a dict object."""
        return isinstance(self.__value, dict)

    def __as_dict(self) -> typing.Dict[str, RawConfigComplexType | '_ConfigImplementation']:
        """Check that inside storage is a dict object. Also adapt code to mypy."""
        self.__ensure_type(dict)
        return typing.cast(typing.Dict[str, RawConfigComplexType | _ConfigImplementation], self.__value)

    def dict_iterate(self) -> typing.Generator[
        typing.Tuple[str, RawConfigComplexType | '_ConfigImplementation'],
        None,
        None
    ]:
        """Return generator that yields keys and values."""
        yield from self.__as_dict().items()

    def dict_properties(self) -> typing.Set[str]:
        """Return keys that this storage holds."""
        return set(self.__as_dict().keys())

    def dict_has(self, name: str) -> bool:
        """Return True if storage has the specified key.

        :param name: a key to check
        """
        return name in self.__as_dict().keys()

    def dict_property(self, name: str) -> '_ConfigImplementation':
        """Return a value that is stored within a key.

        :param name: a key which value should be retrieved
        """
        dict_obj = self.__as_dict()
        if name not in dict_obj:
            raise ValueError(f'There is no {name} property in a configuration object')
        return _ConfigImplementation(dict_obj[name])

    def dict_update(self, name: str, value: RawConfigComplexType | '_ConfigImplementation') -> None:
        """Insert or update a value that is stored within a key.

        :param name: a key which value should be updated
        :param value: new value
        """
        self.__as_dict()[name] = _ConfigImplementation(value).__value

    def dict_clear(self) -> None:
        """Empty internal storage from every key."""
        self.__as_dict().clear()

    def is_list(self) -> bool:
        """Return that inside storage is a list object."""
        return isinstance(self.__value, list)

    def __as_list(self) -> typing.List[RawConfigComplexType | '_ConfigImplementation']:
        """Check that inside storage is a list object. Also adapt code to mypy."""
        self.__ensure_type(list)
        return typing.cast(typing.List[RawConfigComplexType | _ConfigImplementation], self.__value)

    def list_iterate(self) -> typing.Generator['_ConfigImplementation', None, None]:
        """Return generator that yields values from internal list."""
        for i in self.__as_list():
            yield _ConfigImplementation(i)

    def list_get(self, item_index: int) -> '_ConfigImplementation':
        """Get item from internal list by an index.

        :param item_index: index of item to get
        """
        return _ConfigImplementation(self.__as_list()[item_index])

    def is_plain(self) -> bool:
        """Return True if inside storage holds a "plain" object like string or integer."""
        return isinstance(self.__value, (bool, int, float, str))

    def __bool__(self) -> bool:
        """Return value of the internal boolean storage."""
        self.__ensure_type(bool)
        return typing.cast(bool, self.__value)

    def __int__(self) -> int:
        """Return value of the internal integer storage."""
        self.__ensure_type(int)
        return typing.cast(int, self.__value)

    def __float__(self) -> float:
        """Return value of the internal floating-point number storage."""
        self.__ensure_type(float)
        return typing.cast(float, self.__value)

    def __str__(self) -> str:
        """Return value of the internal string storage."""
        self.__ensure_type(str)
        return typing.cast(str, self.__value)


class _ConfigStorage(ConfigStorageProto):
    """This is a shared storage, that is shared between target configuration classes."""

    def __init__(self) -> None:
        """Create a shared storage."""
        self._config = _ConfigImplementation(None)


class Config(_ConfigStorage):
    """Dict-a-like configuration of the application. It works as a main representation of the app config."""

    def __init__(
        self,
        *,
        init_value: typing.Optional[_ConfigImplementation] = None,
        file_obj: typing.Optional[typing.IO[str]] = None,
        property_name: typing.Optional[str] = None
    ):
        """Create a new config.

        :param init_value: initial value
        :param file_obj: optional file object to parse config from. In most cases, if not defined then empty config
        will be created
        :param property_name: optional name of a dict key that will be imported to this configuration from a file.
        If defined then this config will have this key only
        """
        _ConfigStorage.__init__(self)

        self._config = _ConfigImplementation(dict())

        if init_value is not None:
            self.__merge(init_value)

        if file_obj is not None:
            yaml_data = yaml.safe_load(file_obj)  # type: ignore[no-untyped-call]
            if property_name is not None:
                yaml_data = {property_name: yaml_data[property_name]}

            if yaml_data is not None:
                self.__merge(_ConfigImplementation(yaml_data))

    @verify_value(value=lambda x: x.is_dict())
    def __merge(self, value: _ConfigImplementation) -> None:
        """Merge this config with another one.

        :param value: config to merge
        """

        def merge_config(first_config: _ConfigImplementation, second_config: _ConfigImplementation) -> None:
            for property_name, property_value in second_config.dict_iterate():
                impl_property_value = _ConfigImplementation(property_value)
                if not first_config.dict_has(property_name) or not impl_property_value.is_dict():
                    first_config.dict_update(property_name, impl_property_value)
                else:
                    merge_config(first_config.dict_property(property_name), impl_property_value)

        merge_config(self._config, value)

    def properties(self) -> typing.Set[str]:
        """:meth:`.ConfigStorageProto.properties` implementation"""
        return self._config.dict_properties()

    def has_property(self, name: str) -> bool:
        """:meth:`.ConfigStorageProto.has_property` implementation"""
        return self._config.dict_has(name)

    def property(self, name: str) -> ConfigStorageProto:
        """:meth:`.ConfigStorageProto.property` implementation"""
        property_value = self._config.dict_property(name)
        return _cast_implementation(property_value)

    @verify_type(item=str)
    def __getitem__(self, item: typing.Union[str, int]) -> ConfigStorageProto:
        """:meth:`.ConfigStorageProto.__getitem__` implementation"""
        return self.property(item)  # type: ignore[arg-type]  # verify_type do the job

    def reset_properties(self) -> None:
        """:meth:`.ConfigStorageProto.reset_properties` implementation"""
        self._config.dict_clear()

    def merge_file(self, file_obj: typing.IO[str], property_name: typing.Optional[str] = None) -> None:
        """:meth:`.ConfigStorageProto.merge_file` implementation"""
        self.merge_config(Config(file_obj=file_obj, property_name=property_name))

    def merge_config(self, config: 'ConfigStorageProto', property_name: typing.Optional[str] = None) -> None:
        """:meth:`.ConfigStorageProto.merge_config` implementation"""
        if not isinstance(config, Config):
            raise TypeError('Unable to merge with non-Config instance')

        raw_config = config._config
        if property_name is not None:
            if not raw_config.dict_has(property_name):
                raise ValueError(f'There is no {property_name} property in a configuration object')
            raw_config = _ConfigImplementation({property_name: raw_config.dict_property(property_name)})
        self.__merge(raw_config)


class ConfigList(_ConfigStorage):
    """List-a-like configuration of the application."""

    @verify_value(value=lambda x: x is None or x.is_list())
    def __init__(self, value: typing.Optional[_ConfigImplementation] = None) -> None:
        """Create list-a-like config."""
        _ConfigStorage.__init__(self)
        self._config = value if value is not None else _ConfigImplementation([])

    @verify_type(item=int)
    def __getitem__(self, item: typing.Union[int, str]) -> ConfigStorageProto:
        """:meth:`.ConfigStorageProto.__getitem__` implementation"""
        return _cast_implementation(self._config.list_get(item))  # type: ignore[arg-type]  # verify_type do the job

    def iterate_list(self) -> typing.Generator[ConfigStorageProto, None, None]:
        """:meth:`.ConfigStorageProto.iterate_list` implementation"""
        for i in self._config.list_iterate():
            yield _cast_implementation(i)


class ConfigOption(_ConfigStorage):
    """Represents a single option inside a config."""

    @verify_type(value=(None, str, int, float, bool, _ConfigImplementation))
    @verify_value(value=lambda x: x is not isinstance(x, _ConfigImplementation) or x.is_plain() or x.is_null())
    def __init__(self, value: typing.Union[RawConfigPlainTypes, _ConfigImplementation]) -> None:
        """Create a plain option of configuration.

        :param value: initial value
        """
        _ConfigStorage.__init__(self)

        self._config = _ConfigImplementation(value)

    def is_none(self) -> bool:
        """:meth:`.ConfigStorageProto.is_none` implementation"""
        return self._config.is_none()

    def as_str(self) -> str:
        """:meth:`.ConfigStorageProto.as_str` implementation"""
        return str(self._config)

    def as_int(self) -> int:
        """:meth:`.ConfigStorageProto.as_int` implementation"""
        return int(self._config)

    def as_float(self) -> float:
        """:meth:`.ConfigStorageProto.as_float` implementation"""
        return float(self._config)

    def as_bool(self) -> bool:
        """:meth:`.ConfigStorageProto.as_bool` implementation"""
        return bool(self._config)


@verify_type(value=_ConfigImplementation)
def _cast_implementation(value: _ConfigImplementation) -> ConfigStorageProto:
    """Cast internal config representation to "public" objects.

    :param value: internal representation of configuration
    """
    if value.is_none():
        return ConfigOption(None)

    if value.is_dict():
        return Config(init_value=value)

    if value.is_list():
        return ConfigList(value)

    assert(value.is_plain())
    return ConfigOption(value)



# -*- coding: utf-8 -*-
# pyknic/lib/integrated_commands/list_logins.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

import pydantic

from pyknic.lib.bellboy.app import register_bellboy_command, BellBoyCommandHandler
from pyknic.lib.bellboy.models import SecretBackendBellBoyCommandModel
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult, LobbyListValueFeedbackResult


@register_bellboy_command()
class ListLoginsCommand(BellBoyCommandHandler):
    """List known logins
    """

    @classmethod
    def command_name(cls) -> str:
        """The :meth:`.BellBoyCommandHandler.command_name` method implementation"""
        return "list-login"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """The :meth:`.BellBoyCommandHandler.command_model` method implementation"""
        return SecretBackendBellBoyCommandModel

    async def exec(self) -> LobbyCommandResult:
        """The :meth:`.BellBoyInternalCommand.exec_from_cli` method implementation
        """
        assert(isinstance(self._args, SecretBackendBellBoyCommandModel))

        backend = self.secret_backend(self._args.secret_backend)

        all_secrets = backend.get_secrets()
        return LobbyListValueFeedbackResult(
            list_result=[x for x in all_secrets.secrets.keys()]
        )



# -*- coding: utf-8 -*-

import asyncio
import functools
import pytest
import threading
import time
import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SignalsRegistry, SampleTasks

from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.proto import TaskProto, TaskResult, TaskStartError
from pyknic.lib.tasks.threaded_task import ThreadedTask

from fixtures.asyncio import pyknic_async_test


def sample_function(event: threading.Event, start_event: typing.Optional[threading.Event] = None) -> threading.Event:
    if start_event:
        start_event.set()
    event.wait()
    return event


class TestThreadTask:

    def test_plain(self) -> None:
        event = threading.Event()
        task = ThreadedTask.plain_task(functools.partial(sample_function, event))
        task.start()
        event.set()
        task.wait()

    def test_wait_w_timeout(self) -> None:
        event = threading.Event()
        task = ThreadedTask.plain_task(functools.partial(sample_function, event))
        task.start()
        event.set()
        task.wait(100)

    @pytest.mark.parametrize("repeats", range(100))
    @pyknic_async_test
    async def test_start_async(self, repeats: int, module_event_loop: asyncio.AbstractEventLoop) -> None:
        event = threading.Event()
        task = ThreadedTask.plain_task(functools.partial(sample_function, event))

        test_result = 0

        async def async_fn() -> None:
            nonlocal test_result
            for i in range(10):
                test_result += 1
                await asyncio.sleep(0)

        async_test_task = asyncio.create_task(async_fn())
        async_start_task = asyncio.create_task(task.start_async())
        done, pending = await asyncio.wait((async_test_task, async_start_task), return_when=asyncio.FIRST_COMPLETED)

        assert(async_test_task in done)
        assert(async_start_task in pending)

        assert(test_result == 10)
        assert(task.join() is False)
        event.set()

        await async_start_task
        assert(task.join() is True)

    def test_join(self) -> None:
        start_event = threading.Event()
        stop_event = threading.Event()

        task = ThreadedTask.plain_task(functools.partial(sample_function, stop_event, start_event))
        assert(task.join() is True)

        task.start()
        start_event.wait()
        assert(task.join() is False)

        stop_event.set()

        while not task.join():  # this loop is a test itself =)
            time.sleep(0.1)

        assert(task.join() is True)

    def test_start_event(self, signals_registry: 'SignalsRegistry') -> None:
        event = threading.Event()
        task = ThreadedTask.plain_task(functools.partial(sample_function, event))
        task.callback(TaskProto.task_started, signals_registry)

        task.start()
        event.set()
        task.wait()

        assert(signals_registry.dump(True) == [
            (task, TaskProto.task_started, None),
        ])

    def test_fin_event(self, signals_registry: 'SignalsRegistry') -> None:
        plain_task = PlainTask(lambda: None)
        task = ThreadedTask(plain_task)
        task.callback(TaskProto.task_completed, signals_registry)
        task.callback(ThreadedTask.thread_ready, signals_registry)
        task.callback(ThreadedTask.thread_joined, signals_registry)

        task.start()
        task.wait()

        assert(signals_registry.dump(True) == [
            (task, ThreadedTask.thread_ready, plain_task),
            (task, TaskProto.task_completed, TaskResult()),
            (task, ThreadedTask.thread_joined, plain_task)
        ])

    def test_stop(self, sample_tasks: 'SampleTasks') -> None:
        task = ThreadedTask(sample_tasks.LongRunningTask(terminate_method=False))
        task.start()
        task.stop()
        task.wait()

    def test_terminate(self, sample_tasks: 'SampleTasks') -> None:
        task = ThreadedTask(sample_tasks.LongRunningTask(stop_method=False))
        task.start()
        task.terminate()
        task.wait()

    def test_start_exception(self) -> None:
        event = threading.Event()
        task = ThreadedTask.plain_task(functools.partial(sample_function, event))
        task.start()

        with pytest.raises(TaskStartError):
            task.start()

        event.set()
        task.wait()

    def test_task_result(self, signals_registry: 'SignalsRegistry') -> None:
        event = threading.Event()
        task = ThreadedTask.plain_task(functools.partial(sample_function, event))
        task.callback(TaskProto.task_completed, signals_registry)

        task.start()
        event.set()
        task.wait()

        assert(signals_registry.dump(True) == [
            (task, TaskProto.task_completed, TaskResult()),
        ])

    def test_task_exception(self, signals_registry: 'SignalsRegistry') -> None:
        class Task(TaskProto):

            def __init__(self, exc: BaseException) -> None:
                TaskProto.__init__(self)
                self.__exc = exc

            def start(self) -> None:
                raise self.__exc

        task_exception = ValueError('!')
        original_task = Task(task_exception)
        task = ThreadedTask(original_task)
        signals_registry.thread_safe()
        task.callback(TaskProto.task_completed, signals_registry)
        task.callback(ThreadedTask.thread_ready, signals_registry)

        task.start()
        task.wait()

        assert(signals_registry.dump(True) == [
            (task, ThreadedTask.thread_ready, original_task),
            (task, TaskProto.task_completed, TaskResult(exception=task_exception)),
        ])



# -*- coding: utf-8 -*-
# pyknic/tasks/log.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from pyknic.lib.tasks.scheduler.chain_source import ChainedTask, __default_chained_tasks_registry__
from pyknic.lib.registry import register_api
from pyknic.lib.log import Logger


@register_api(__default_chained_tasks_registry__, "log_task")
class LogTask(ChainedTask):
    """ This task sets up the project's logger
    """

    __env_var_name__ = "PYKNIC_LOG_LEVEL"  # defines an environment variable for log level

    def start(self) -> None:
        """ Call the method :meth:`.LogTask.setup_log` to tweak the logger
        """
        self.setup_log()

    def task_name(self) -> str:
        """ Return this task's name
        """
        return "logger-setter"

    @staticmethod
    def setup_log() -> None:
        """ Tweak the logger
        """
        if LogTask.__env_var_name__ not in os.environ:
            raise ValueError(f'The "{LogTask.__env_var_name__}" variable was not set')

        log_level_str = os.environ[LogTask.__env_var_name__]
        if log_level_str not in ('ERROR', 'WARN', 'INFO', 'DEBUG'):
            raise ValueError(f'Unknown value of "{LogTask.__env_var_name__}" -- {log_level_str}')

        formatter = logging.Formatter(
            '[%(name)s] [%(threadName)s] [%(levelname)s] [%(asctime)s] %(message)s',
            '%Y-%m-%d %H:%M:%S'
        )

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)

        log_level = getattr(logging, log_level_str)
        Logger.addHandler(handler)
        Logger.setLevel(log_level)

        Logger.info('Logger was set up')



# -*- coding: utf-8 -*-

from pyknic.lib.x_mansion import CapabilitiesAndSignalsMeta, CapabilitiesAndSignals

from pyknic.lib.signals.proto import Signal
from pyknic.lib.signals.source import SignalSourceMeta, SignalSource
from pyknic.lib.capability import CapabilitiesHolderMeta, CapabilitiesHolder, capability, iscapable


class TestCapabilitiesAndSignalsMeta:

    class Sample(metaclass=CapabilitiesAndSignalsMeta):
        pass

    def test(self) -> None:
        assert(isinstance(TestCapabilitiesAndSignalsMeta.Sample, SignalSourceMeta))
        assert(isinstance(TestCapabilitiesAndSignalsMeta.Sample, CapabilitiesHolderMeta))


class TestCapabilitiesAndSignals:

    class BaseCls(CapabilitiesAndSignals):
        signal = Signal()

        @capability
        def foo(self) -> None:
            pass

    class DerivedCls(BaseCls):

        def foo(self) -> None:
            pass

    def test(self) -> None:
        base_cls = TestCapabilitiesAndSignals.BaseCls()

        assert(isinstance(TestCapabilitiesAndSignals.BaseCls, CapabilitiesAndSignalsMeta))
        assert(isinstance(base_cls, SignalSource))
        assert(isinstance(base_cls, CapabilitiesHolder))

        base_cls.emit(base_cls.signal)
        assert(iscapable(base_cls, TestCapabilitiesAndSignals.BaseCls.foo) is False)
        assert(iscapable(TestCapabilitiesAndSignals.DerivedCls(), TestCapabilitiesAndSignals.BaseCls.foo) is True)



# -*- coding: utf-8 -*-

import pytest
import typing

from datetime import datetime, timezone

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SignalsRegistry, SampleTasks

from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.scheduler.cron_source import CronSchedule, CronRecordInitError, CronScheduleRecord, CronTaskSource


def test_exceptions() -> None:
    assert(issubclass(CronRecordInitError, Exception) is True)


class TestCronSchedule:

    iter_init_date = datetime(year=2016, month=10, day=27, hour=22, minute=59, second=10, tzinfo=timezone.utc)

    @pytest.mark.parametrize(
        "cron_schedule, results", [
            (
                "20 20 10 * 10",
                (
                    datetime(year=2017, month=10, day=10, hour=20, minute=20, second=0, tzinfo=timezone.utc),
                    datetime(year=2018, month=10, day=10, hour=20, minute=20, second=0, tzinfo=timezone.utc),
                    datetime(year=2019, month=10, day=10, hour=20, minute=20, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "* * * * *",
                (
                    datetime(year=2016, month=10, day=27, hour=22, minute=59, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=27, hour=23, minute=0, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=27, hour=23, minute=1, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "10 * * * *",
                (
                    datetime(year=2016, month=10, day=27, hour=23, minute=10, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=28, hour=0, minute=10, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=28, hour=1, minute=10, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "* 10 * * *",
                (
                    datetime(year=2016, month=10, day=28, hour=10, minute=0, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=28, hour=10, minute=1, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=28, hour=10, minute=2, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "* * 10 * *",
                (
                    datetime(year=2016, month=11, day=10, hour=0, minute=0, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=10, hour=0, minute=1, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=10, hour=0, minute=2, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "* * * 2 *",
                (
                    datetime(year=2016, month=11, day=1, hour=0, minute=0, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=1, hour=0, minute=1, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=1, hour=0, minute=2, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "* * * * 6",
                (
                    datetime(year=2017, month=6, day=1, hour=0, minute=0, second=0, tzinfo=timezone.utc),
                    datetime(year=2017, month=6, day=1, hour=0, minute=1, second=0, tzinfo=timezone.utc),
                    datetime(year=2017, month=6, day=1, hour=0, minute=2, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "15 13 11 * *",
                (
                    datetime(year=2016, month=11, day=11, hour=13, minute=15, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=12, day=11, hour=13, minute=15, second=0, tzinfo=timezone.utc),
                    datetime(year=2017, month=1, day=11, hour=13, minute=15, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "15 13 * 2 *",
                (
                    datetime(year=2016, month=11, day=1, hour=13, minute=15, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=8, hour=13, minute=15, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=15, hour=13, minute=15, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "7 7 * * *",
                (
                    datetime(year=2016, month=10, day=28, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=29, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=30, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=10, day=31, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=1, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2016, month=11, day=2, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                )
            ),

            (
                "7 7 7 7 7",
                (
                    datetime(year=2019, month=7, day=7, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2024, month=7, day=7, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2030, month=7, day=7, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                    datetime(year=2041, month=7, day=7, hour=7, minute=7, second=0, tzinfo=timezone.utc),
                )
            ),

        ]
    )
    def test(self, cron_schedule: str, results: typing.Tuple[datetime]) -> None:
        schedule = CronSchedule.from_string(cron_schedule)
        cron_iterator = schedule.iterate(TestCronSchedule.iter_init_date)
        for r in results:
            assert(next(cron_iterator) == r)

    def test_exception(self) -> None:
        schedule = CronSchedule.from_string("* * * * *")

        with pytest.raises(ValueError):
            schedule.iterate(datetime.now())  # doesn't have tzinfo -- exception is raised

        with pytest.raises(ValueError):
            CronSchedule.from_string("* * 31 * 2")  # there is no 31th of Februrary

        with pytest.raises(ValueError):
            CronSchedule.from_string("* *")  # tokens are not enough

        with pytest.raises(ValueError):
            CronSchedule.from_string("* * * * * *")  # too many tokens


class TestCronScheduleRecord:

    def test_plain(self, sample_tasks: 'SampleTasks') -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        schedule = CronSchedule.from_string("* * * * *")
        cron_record = CronScheduleRecord(record, schedule)

        assert(cron_record.record() is record)

    @pytest.mark.parametrize(
        "method_name", [
            "next",
            "commit"
        ]
    )
    def test_uninitialized_record(self, sample_tasks: 'SampleTasks', method_name: str) -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        schedule = CronSchedule.from_string("* * * * *")
        cron_record = CronScheduleRecord(record, schedule)

        with pytest.raises(CronRecordInitError):
            getattr(cron_record, method_name)()

    def test_double_init(self, sample_tasks: 'SampleTasks') -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        schedule = CronSchedule.from_string("* * * * *")
        cron_record = CronScheduleRecord(record, schedule)

        cron_record.init(datetime.now(timezone.utc))
        with pytest.raises(CronRecordInitError):
            cron_record.init(datetime.now(timezone.utc))

    def test_next(self, sample_tasks: 'SampleTasks') -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        schedule = CronSchedule.from_string("* * * * *")
        cron_record = CronScheduleRecord(record, schedule)
        start_datetime = datetime(year=2001, month=1, day=1, hour=1, minute=1, second=0, tzinfo=timezone.utc)

        cron_record.init(start_datetime)
        next_datetime = cron_record.next()
        assert(next_datetime == datetime(year=2001, month=1, day=1, hour=1, minute=1, second=0, tzinfo=timezone.utc))

        next_datetime = cron_record.next()
        assert(next_datetime == datetime(year=2001, month=1, day=1, hour=1, minute=1, second=0, tzinfo=timezone.utc))

        cron_record.commit()
        next_datetime = cron_record.next()
        assert(next_datetime == datetime(year=2001, month=1, day=1, hour=1, minute=2, second=0, tzinfo=timezone.utc))

        next_datetime = cron_record.next()
        assert(next_datetime == datetime(year=2001, month=1, day=1, hour=1, minute=2, second=0, tzinfo=timezone.utc))

        cron_record.commit()
        next_datetime = cron_record.next()
        assert(next_datetime == datetime(year=2001, month=1, day=1, hour=1, minute=3, second=0, tzinfo=timezone.utc))

        next_datetime = cron_record.next()
        assert(next_datetime == datetime(year=2001, month=1, day=1, hour=1, minute=3, second=0, tzinfo=timezone.utc))

    def test_schedule(self, sample_tasks: 'SampleTasks') -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        schedule = CronSchedule.from_string("* * * * *")
        cron_record = CronScheduleRecord(record, schedule)
        start_datetime = datetime(year=2001, month=1, day=1, hour=1, minute=1, second=0, tzinfo=timezone.utc)

        schedule_obj = cron_record.schedule(start_datetime)

        assert(
            next(schedule_obj) == datetime(year=2001, month=1, day=1, hour=1, minute=1, second=0, tzinfo=timezone.utc)
        )
        assert(
            next(schedule_obj) == datetime(year=2001, month=1, day=1, hour=1, minute=2, second=0, tzinfo=timezone.utc)
        )
        assert(
            next(schedule_obj) == datetime(year=2001, month=1, day=1, hour=1, minute=3, second=0, tzinfo=timezone.utc)
        )
        assert(
            next(schedule_obj) == datetime(year=2001, month=1, day=1, hour=1, minute=4, second=0, tzinfo=timezone.utc)
        )


class TestCronTaskSource:

    def test(self) -> None:
        source = CronTaskSource()
        assert(source.poll() is None)
        assert(list(source.records()) == [])

    def test_polling(self, sample_tasks: 'SampleTasks', signals_registry: 'SignalsRegistry') -> None:
        source = CronTaskSource()
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        cron_schedule = CronScheduleRecord(record, CronSchedule.from_string("* * * * *"))

        source.callback(CronTaskSource.task_scheduled, signals_registry)
        source.submit_record(cron_schedule)
        source.poll()

        assert(signals_registry.dump(True) == [
            (source, CronTaskSource.task_scheduled, record),
        ])

        source.poll()
        assert(signals_registry.dump(True) == [])

    def test_records(self, sample_tasks: 'SampleTasks', signals_registry: 'SignalsRegistry') -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        now = datetime.now(timezone.utc)

        cron_schedule1 = CronScheduleRecord(record, CronSchedule.from_string(f"* {(now.hour + 4) % 24} * * *"))
        cron_schedule2 = CronScheduleRecord(record, CronSchedule.from_string(f"* {(now.hour + 3) % 24} * * *"))
        cron_schedule3 = CronScheduleRecord(record, CronSchedule.from_string(f"* {(now.hour + 5) % 24} * * *"))

        source = CronTaskSource()
        source.callback(CronTaskSource.polling_update, signals_registry)

        source.submit_record(cron_schedule1)
        signals = signals_registry.dump(True)
        ttl1 = signals[0][2]
        assert(len(signals) == 1)
        assert(signals[0][1] == CronTaskSource.polling_update)
        assert(ttl1 > 0)  # type: ignore[operator]  # it's a test

        source.submit_record(cron_schedule2)
        signals = signals_registry.dump(True)
        ttl2 = signals[0][2]
        assert(len(signals) == 1)
        assert(signals[0][1] == CronTaskSource.polling_update)
        assert(ttl2 > 0)  # type: ignore[operator]  # it's a test
        assert(ttl2 < ttl1)  # type: ignore[operator]  # it's a test

        source.submit_record(cron_schedule3)
        assert(signals_registry.dump(True) == [])  # no polling update is required

        assert(list(source.records()) == [cron_schedule2, cron_schedule1, cron_schedule3])

    def test_discard(self, sample_tasks: 'SampleTasks', signals_registry: 'SignalsRegistry') -> None:
        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        now = datetime.now(timezone.utc)

        cron_schedule1 = CronScheduleRecord(record, CronSchedule.from_string(f"* {(now.hour + 4) % 24} * * *"))
        cron_schedule2 = CronScheduleRecord(record, CronSchedule.from_string(f"* {(now.hour + 3) % 24} * * *"))
        cron_schedule3 = CronScheduleRecord(record, CronSchedule.from_string(f"* {(now.hour + 5) % 24} * * *"))

        source = CronTaskSource()
        source.submit_record(cron_schedule1)
        source.submit_record(cron_schedule2)
        source.submit_record(cron_schedule3)

        source.callback(CronTaskSource.polling_update, signals_registry)

        source.discard_record(cron_schedule2)
        signals = signals_registry.dump(True)
        ttl1 = signals[0][2]
        assert(len(signals) == 1)
        assert(signals[0][1] == CronTaskSource.polling_update)
        assert(ttl1 > 0)  # type: ignore[operator]  # it's a test

        source.discard_record(cron_schedule1)
        signals = signals_registry.dump(True)
        ttl2 = signals[0][2]
        assert(len(signals) == 1)
        assert(signals[0][1] == CronTaskSource.polling_update)
        assert(ttl2 > 0)  # type: ignore[operator]  # it's a test
        assert(ttl2 < ttl1)  # type: ignore[operator]  # it's a test

        assert(list(source.records()) == [cron_schedule3])

        with pytest.raises(ValueError):
            source.discard_record(cron_schedule1)

        source.discard_record(cron_schedule3)  # no mo records left
        assert(signals_registry.dump(True) == [])



# -*- coding: utf-8 -*-

import asyncio
import pytest
import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SampleTasks

from pyknic.lib.capability import iscapable, CapabilityDescriptor

from pyknic.lib.tasks.proto import TaskStartError, TaskStopError, NoSuchTaskError, TaskProto
from pyknic.lib.tasks.proto import TaskResult, ScheduleRecordProto, SchedulerProto, ScheduledTaskPostponePolicy
from pyknic.lib.tasks.proto import TaskExecutorProto, ScheduleSourceProto

from fixtures.asyncio import pyknic_async_test


def test_exceptions() -> None:
    assert(issubclass(TaskStartError, Exception) is True)
    assert(issubclass(TaskStopError, Exception) is True)
    assert(issubclass(NoSuchTaskError, Exception) is True)


@pyknic_async_test
async def test_abstract(module_event_loop: asyncio.AbstractEventLoop) -> None:
    pytest.raises(TypeError, TaskProto)
    pytest.raises(NotImplementedError, TaskProto.start, None)

    pytest.raises(TypeError, ScheduleRecordProto)
    pytest.raises(NotImplementedError, ScheduleRecordProto.task, None)
    pytest.raises(NotImplementedError, ScheduleRecordProto.source, None)

    pytest.raises(TypeError, SchedulerProto)
    pytest.raises(NotImplementedError, SchedulerProto.subscribe, None, None)
    pytest.raises(NotImplementedError, SchedulerProto.unsubscribe, None, None)

    pytest.raises(NotImplementedError, ScheduleSourceProto.scheduler_feedback, None, None, None)  # capability test

    pytest.raises(TypeError, TaskExecutorProto)
    pytest.raises(NotImplementedError, TaskExecutorProto.submit_task, None, None)
    pytest.raises(NotImplementedError, TaskExecutorProto.complete_task, None, None)
    pytest.raises(NotImplementedError, TaskExecutorProto.wait_task, None, None, None)

    with pytest.raises(NotImplementedError):
        await TaskExecutorProto.start_async(None, None)  # type: ignore[arg-type]  # it is just a test

    pytest.raises(NotImplementedError, TaskExecutorProto.tasks, None)


class TestTaskProto:

    def test(self, sample_tasks: 'SampleTasks') -> None:
        assert(isinstance(
            TaskProto.stop.__pyknic_capability__,  # type: ignore[attr-defined]  # mypy and metaclass issues
            CapabilityDescriptor
        ))
        assert(isinstance(
            TaskProto.terminate.__pyknic_capability__,  # type: ignore[attr-defined]  # mypy and metaclass issues
            CapabilityDescriptor
        ))

        task = sample_tasks.DummyTask()
        pytest.raises(NotImplementedError, task.stop)
        pytest.raises(NotImplementedError, task.terminate)

        assert(iscapable(task, TaskProto.stop) is False)
        assert(iscapable(task, TaskProto.terminate) is False)

        task.emit(TaskProto.task_started)
        task.emit(TaskProto.task_completed, TaskResult())

        assert(TaskProto.task_name(None) is None)  # type: ignore[arg-type]


class TestScheduleRecordProto:

    def test(self, sample_tasks: 'SampleTasks') -> None:

        class Record(ScheduleRecordProto):
            def task(self) -> TaskProto:
                return sample_tasks.DummyTask()

            def source(self) -> 'ScheduleSourceProto':
                return object()  # type: ignore[return-value]  # just a test

        record = Record()
        assert(record.group_id() is None)
        assert(record.ttl() is None)
        assert(record.simultaneous_runs() == 0)
        assert(record.postpone_policy() == ScheduledTaskPostponePolicy.wait)



# -*- coding: utf-8 -*-

from fixtures.callbacks_n_signals import callbacks_registry, signals_registry, signal_watcher
from fixtures.callbacks_n_signals import CallbackRegistry, SignalsRegistry, SignalWatcher
from fixtures.config import config_file
from fixtures.event_loop import event_loop, class_event_loop, module_event_loop, package_event_loop, session_event_loop
from fixtures.gettext import gettext
from fixtures.fastapi import fastapi_fixture, fastapi_class_fixture, fastapi_module_fixture, fastapi_package_fixture
from fixtures.fastapi import fastapi_session_fixture, AsyncFastAPIFixture
from fixtures.lobby import lobby_shm_secrets
from fixtures.sftp_server import sftp_server
from fixtures.tasks import sample_tasks, empty_datalog, SampleTasks
from fixtures.tgbot import tgbot_fixture, TGBotFixture

__all__ = [
    "callbacks_registry",
    "config_file",
    "empty_datalog",

    "event_loop",
    "class_event_loop",
    "module_event_loop",
    "package_event_loop",
    "session_event_loop",

    "fastapi_fixture",
    "fastapi_class_fixture",
    "fastapi_module_fixture",
    "fastapi_package_fixture",
    "fastapi_session_fixture",

    "gettext",

    "sample_tasks",
    "signals_registry",
    "signal_watcher",

    "AsyncFastAPIFixture",
    "CallbackRegistry",
    "SampleTasks",
    "SignalsRegistry",
    "SignalWatcher",

    "tgbot_fixture",
    "TGBotFixture",

    "sftp_server",

    "lobby_shm_secrets"
]



# -*- coding: utf-8 -*-

import aiohttp
import asyncio
import fastapi
import json
import pydantic_core
import pytest
import typing

from fixtures.asyncio import pyknic_async_test

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import AsyncFastAPIFixture

from pyknic.lib.fastapi.base import BaseFastAPIApp, TgBotBaseFastAPIApp
from pyknic.lib.fastapi.models import tg_bot_types, tg_bot_methods, base as base_models
from pyknic.lib.config import Config
from pyknic.lib.gettext import GetTextWrapper
from pyknic.path import root_path


class TestBaseFastAPIApp:

    @pyknic_async_test
    async def test_abstract(
        self,
        module_event_loop: asyncio.AbstractEventLoop,
        fastapi_module_fixture: 'AsyncFastAPIFixture',
        gettext: GetTextWrapper
    ) -> None:
        pytest.raises(TypeError, BaseFastAPIApp, Config(), gettext)
        pytest.raises(NotImplementedError, BaseFastAPIApp.create_app, fastapi_module_fixture.fastapi, Config(), gettext)

    @pyknic_async_test
    async def test_sample(
        self,
        module_event_loop: asyncio.AbstractEventLoop,
        fastapi_module_fixture: 'AsyncFastAPIFixture',
        gettext: GetTextWrapper
    ) -> None:

        class SampleApp(BaseFastAPIApp):

            async def root(self, req: fastapi.Request) -> typing.Dict[str, int]:
                return {"foo": 1}

            @classmethod
            def create_app(
                cls, fastapi_app: fastapi.FastAPI, config: Config, translations: GetTextWrapper
            ) -> typing.Any:
                app = cls(config, translations)

                fastapi_app.get(
                    '/sample/app',
                    status_code=200,
                )(app.root)

                return app

        config = Config()
        app = SampleApp.create_app(fastapi_module_fixture.fastapi, config, gettext)

        assert(app.config() is config)
        assert(app.lang('en').gettext("Choose a game") == "Choose a game")
        assert(app.lang('fr').gettext("Choose a game") == "Choose a game")
        assert(app.lang('ru').gettext("Choose a game") == "Выбирай игру")

        async with aiohttp.request('get', 'http://localhost:8000/unknow/path') as response:
            assert(response.status == 404)

        async with aiohttp.request('get', 'http://localhost:8000/sample/app') as response:
            assert(response.status == 200)
            data = await response.text()
            assert(json.loads(data) == {"foo": 1})


class TestTgBotBaseFastAPIApp:

    @pyknic_async_test
    async def test_abstract(
        self,
        module_event_loop: asyncio.AbstractEventLoop,
        fastapi_module_fixture: 'AsyncFastAPIFixture',
        gettext: GetTextWrapper
    ) -> None:
        pytest.raises(TypeError, TgBotBaseFastAPIApp, Config(), gettext)
        pytest.raises(NotImplementedError, TgBotBaseFastAPIApp.bot_path, Config())

    @pyknic_async_test
    async def test_sample(
        self,
        module_event_loop: asyncio.AbstractEventLoop,
        fastapi_module_fixture: 'AsyncFastAPIFixture',
        gettext: GetTextWrapper
    ) -> None:

        class SampleBot(TgBotBaseFastAPIApp):

            @classmethod
            def bot_path(cls, config: Config) -> str:
                return '/smart/bot'

            async def process_message(
                self, tg_update: tg_bot_types.Update
            ) -> tg_bot_methods.MethodSendMessage | base_models.NullableModel | None:
                assert(tg_update.message)
                assert(tg_update.message.text)

                if tg_update.message.text != 'Hello!':
                    return self.reply(tg_update, tg_update.message.text)
                return None

        config = Config()
        translations = GetTextWrapper(root_path / 'locales')
        app = SampleBot.create_app(fastapi_module_fixture.fastapi, config, translations)

        request1 = tg_bot_types.Update(
            update_id=1,
            message=tg_bot_types.Message(  # type: ignore[call-arg]
                message_id=20,
                from_=tg_bot_types.User(id_=300),  # type: ignore[call-arg]
                chat=tg_bot_types.Chat(id_=4000, type_="private"),  # type: ignore[call-arg]
                text="Hello!"
            )
        )

        async with (aiohttp.request(
            'post',
            'http://localhost:8000/smart/bot',
            json=request1.model_dump(exclude_none=True, by_alias=True)
        ) as response):
            assert(response.status == 200)
            data = await response.text()
            assert(json.loads(data) == dict())

        request2 = tg_bot_types.Update(
            update_id=2,
            message=tg_bot_types.Message(  # type: ignore[call-arg]
                message_id=30,
                from_=tg_bot_types.User(id_=400),  # type: ignore[call-arg]
                chat=tg_bot_types.Chat(id_=5000, type_="private"),  # type: ignore[call-arg]
                text="Who are you?"
            )
        )

        async with (aiohttp.request(
            'post',
            'http://localhost:8000/smart/bot',
            json=request2.model_dump(exclude_none=True, by_alias=True)
        ) as response):
            assert(response.status == 200)
            data = await response.text()
            assert(json.loads(data)['method'] == "sendMessage")  # defaults check

            bot_response = tg_bot_methods.MethodSendMessage.model_validate(
                pydantic_core.from_json(data, allow_partial=True)
            )
            assert(bot_response.method == "sendMessage")
            assert(bot_response.reply_parameters is not None)
            assert(bot_response.reply_parameters.message_id == 30)
            assert(bot_response.text == 'Who are you?')

        user1 = tg_bot_types.User(id_=500)  # type: ignore[call-arg]
        assert(app.user_lang(user1).gettext("Choose a game") == "Choose a game")
        user2 = tg_bot_types.User(id_=500, language_code='fr')  # type: ignore[call-arg]
        assert(app.user_lang(user2).gettext("Choose a game") == "Choose a game")
        user3 = tg_bot_types.User(id_=500, language_code='ru')  # type: ignore[call-arg]
        assert(app.user_lang(user3).gettext("Choose a game") == "Выбирай игру")



# -*- coding: utf-8 -*-

import pytest
import typing

from abc import ABCMeta

from pyknic.lib.singleton import SingletonMeta, create_singleton


class TestSingletonMeta:

    def test(self) -> None:

        class BaseClass:
            def __init__(self, x: int) -> None:
                self.x = x

            def pow(self, y: int) -> typing.Any:
                return self.x ** y

        class SingletonClass(metaclass=SingletonMeta):
            pass

        value = BaseClass(5)
        assert(SingletonClass.singleton() is None)
        SingletonClass.setup_singleton(value)
        assert(SingletonClass.singleton() is value)

        with pytest.raises(ValueError):
            SingletonClass.setup_singleton(BaseClass(6))

        assert(SingletonClass.x == 5)
        assert(SingletonClass.pow(2) == 25)

        SingletonClass.x = 7
        assert(SingletonClass.pow(2) == 49)


def test_create_singleton() -> None:
    class BaseClass:
        pass

    value = BaseClass()
    singleton = create_singleton(value)
    assert(isinstance(singleton, SingletonMeta))
    assert(ABCMeta.__getattribute__(singleton, "__instance__") is value)



# -*- coding: utf-8 -*-
# pyknic/lib/crypto/random.py
#
# Copyright (C) 2016-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import math
import os
import sys

from pyknic.lib.verify import verify_value


"""This module generates random data secure enough for cryptography purpose.

see also: https://cryptography.io/en/latest/random-numbers/
"""


@verify_value(bits_count=lambda x: x > 0)
def random_bits(bits_count: int) -> int:
    """Random generator that produces the specified number of random bits. The result is a non-negative integer.

    :param bits_count: random bits to generate
    :return: int
    """
    bytes_count = int(math.ceil(bits_count / 8))
    random_value = int.from_bytes(os.urandom(bytes_count), byteorder=sys.byteorder)
    result_bits = bytes_count * 8

    if result_bits > bits_count:
        random_value = (random_value >> (result_bits - bits_count))

    return random_value


@verify_value(maximum_value=lambda x: x > 0)
def random_int(maximum_value: int) -> int:
    """Random generator that produces a single integer that resides between 0 and the "maximum_value".
    The result is a non-negative value.

    :param maximum_value: maximum integer value
    :return: int
    """

    if maximum_value == 1:
        return random_bits(1)

    bits = math.floor(math.log2(maximum_value))
    result = random_bits(bits) + random_int(maximum_value - ((2 ** bits) - 1))
    return result


@verify_value(bytes_count=lambda x: x >= 0)
def random_bytes(bytes_count: int) -> bytes:
    """Generate random bytes sequence

    :param bytes_count: sequence length
    :return: bytes
    """
    return bytes([random_bits(8) for x in range(bytes_count)])
